############################################################################
# tools/pynuttx/nxgdb/memdump.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import binascii
import os
import re
from collections import defaultdict
from typing import Dict, Generator, List, Protocol, Tuple

import gdb

from . import autocompeletion, backtrace, mm, utils


class MMNodeDump(Protocol):
    """Node information protocol for dump"""

    address: int  # Note that address should be in type of int
    nodesize: int
    seqno: int
    pid: int
    backtrace: Tuple[int]
    is_free: bool
    from_pool: bool
    overhead: int
    useraddress: int  # Address of user memory
    usersize: int = 0  # Size of the user data, excluding overhead

    def contains(self, addr: int) -> bool: ...

    def read_memory(self) -> memoryview: ...


def filter_node(
    pid=None,
    nodesize=None,
    used=None,
    free=None,
    seqmin=None,
    seqmax=None,
    orphan=None,
    no_pid=None,
    no_heap=None,
    no_pool=None,
) -> bool:
    return lambda node: (
        (pid is None or node.pid == pid)
        and (no_pid is None or node.pid != no_pid)
        and (nodesize is None or node.nodesize == nodesize)
        and (not used or not node.is_free)
        and (not free or node.is_free)
        and (seqmin is None or node.seqno >= seqmin)
        and (seqmax is None or node.seqno <= seqmax)
        and (not orphan or node.is_orphan)
    )


def dump_nodes(
    heaps: List[mm.MMHeap],
    filters=None,
) -> Generator[MMNodeDump, None, None]:
    no_heap = filters and filters.get("no_heap")
    no_pool = filters and filters.get("no_pool")

    if not no_heap:
        yield from (
            node
            for heap in heaps
            for node in filter(filter_node(**filters), heap.nodes)
        )

    if not no_pool:
        yield from (
            blk
            for pool in mm.get_pools(heaps)
            for blk in filter(filter_node(**filters), pool.blks)
        )


def group_nodes(
    nodes: List[MMNodeDump], grouped: Dict[MMNodeDump, List[MMNodeDump]] = None
) -> Dict[MMNodeDump, List[MMNodeDump]]:
    grouped = grouped or defaultdict(list)
    for node in nodes:
        # default to group by same PID, same size and same backtrace
        grouped[node].append(node)
    return grouped


def print_node(node: MMNodeDump, alive, count=1, formatter=None, no_backtrace=False):
    formatter = (
        formatter or "{:>1} {:>4} {:>12} {:>12} {:>12} {:>9} {:>14} {:>18} {:}\n"
    )
    gdb.write(
        formatter.format(
            "\x1b[33;1m*\x1b[m" if not alive else "",
            "P" if node.from_pool else "H",
            count,
            node.pid,
            node.nodesize,
            node.overhead,
            node.seqno,
            hex(node.address),
            "",
        )
    )

    if mm.MM_RECORD_STACK_DEPTH > 0 and not no_backtrace:
        leading = formatter.format("", "", "", "", "", "", "", "", "")[:-1]
        btformat = leading + "{1:<48}{2}\n"
        if node.backtrace and node.backtrace[0]:
            gdb.write(f"{backtrace.Backtrace(node.backtrace, formatter=btformat)}\n")


def print_header(formatter=None):
    formatter = (
        formatter or "{:>1} {:>4} {:>12} {:>12} {:>12} {:>9} {:>14} {:>18} {:}\n"
    )
    head = (
        "",
        "Pool",
        "CNT",
        "PID",
        "NodeSize",
        "Overhead",
        "Seqno",
        "Address",
        "Backtrace",
    )
    gdb.write(formatter.format(*head))


def get_heaps(args_heap: str = None) -> List[mm.MMHeap]:
    """Get the list of heaps, or a specific heap if args_heap is provided."""
    if args_heap is not None:
        try:
            # Check if the arg is heap name, instead of heap symbol or address.
            heaps = mm.get_heaps()
            heap = next((heap for heap in heaps if heap.name == args_heap), None)
            if heap is not None:
                return [heap]
        except gdb.MemoryError:
            pass
        finally:
            return [mm.MMHeap(utils.parse_arg(args_heap))]
    else:
        return mm.get_heaps()


def parse_memdump_log(logfile, filters=None) -> Generator[MMNodeDump, None, None]:
    nodes = []

    class DumpNode(MMNodeDump):
        def __init__(self, address, nodesize, seqno, pid, backtrace, is_free, overhead):
            # C code dump the start address of the node, convert it the actual start address
            self.address = address - overhead
            self.nodesize = nodesize
            self.seqno = seqno
            self.pid = pid
            self.backtrace = backtrace
            self.overhead = overhead
            self.is_free = False
            self.from_pool = False

        def __repr__(self) -> str:
            return f"node@{self.address:#x}: size:{self.nodesize} seq:{self.seqno} pid:{self.pid} "

        def contains(self, addr: int) -> bool:
            return self.address <= addr < self.address + self.nodesize

        @property
        def prevnode(self):
            return next(
                (node for node in nodes if node.contains(self.address - 1)), None
            )

        @property
        def nextnode(self):
            address = self.address + self.nodesize  # address of the next node
            return next(
                (node for node in nodes if node.address == address),
                None,
            )

    with open(logfile, "r", errors="ignore") as f:
        for line in f:
            match = re.search(
                r"(\d+)\s+(\d+)\s+(\d+)\s+(\d+)((?:\s+0x[0-9a-fA-F]+)+)", line
            )
            if not match:
                continue

            try:
                pid = int(match.group(1))
                size = int(match.group(2))
                overhead = int(match.group(3))
                seq = int(match.group(4))
                addresses = match.group(5).split()
                addr = int(addresses[0], base=16)
                mem = tuple(int(addr, base=16) for addr in addresses[1:])
                nodes.append(DumpNode(addr, size, seq, pid, mem, False, overhead))
            except Exception as e:
                print(f"Error parsing line: {line}, {e}")

    return filter(filter_node(**filters), nodes) if filters else nodes


@autocompeletion.complete
class MMDump(gdb.Command):
    """Dump memory manager heap"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-a",
            "--address",
            type=str,
            metavar="symbol",
            default=None,
            help="Find the node that contains the address and exit",
        )

        parser.add_argument(
            "-l",
            "--log",
            type=str,
            metavar="file",
            default=None,
            help="Use the memdump log file generated by memdump command on device instead of live dump",
        )

        parser.add_argument(
            "--heap",
            type=str,
            metavar="symbol",
            default=None,
            help="Which heap to inspect",
        )

        parser.add_argument(
            "-p", "--pid", type=int, default=None, help="Thread PID, -1 for mempool"
        )
        parser.add_argument(
            "-i", "--min", type=int, default=None, help="Minimum sequence number"
        )
        parser.add_argument(
            "-x", "--max", type=int, default=None, help="Maximum sequence number"
        )
        parser.add_argument("--free", action="store_true", help="Free flag")
        parser.add_argument("--biggest", action="store_true", help="biggest allocated")
        parser.add_argument(
            "--orphan", action="store_true", help="Filter nodes that are orphan"
        )
        parser.add_argument(
            "--top", type=int, default=None, help="biggest top n, default to all"
        )
        parser.add_argument(
            "--size", type=int, default=None, help="Node block size filter."
        )
        parser.add_argument(
            "--no-pool",
            "--nop",
            action="store_true",
            help="Exclude dump from memory pool",
        )
        parser.add_argument(
            "--no-heap", "--noh", action="store_true", help="Exclude heap dump"
        )
        parser.add_argument(
            "--no-group", "--nog", action="store_true", help="Do not group the nodes"
        )
        parser.add_argument(
            "--no-backtrace",
            "--nob",
            action="store_true",
            help="Do not print backtrace",
        )
        parser.add_argument(
            "--no-reverse",
            "--nor",
            action="store_true",
            help="Do not reverse the sort result",
        )
        parser.add_argument(
            "--no-pid", type=int, default=None, help="Exclude nodes from this PID"
        )

        # add option to sort the node by size or count
        parser.add_argument(
            "--sort",
            type=str,
            choices=["size", "nodesize", "count", "seq", "address"],
            default="count",
            help="sort the node by size(nodesize * count), nodesize,  count or sequence number",
        )
        return parser

    def __init__(self):
        super().__init__("mm dump", gdb.COMMAND_USER)
        # define memdump as mm dump
        utils.alias("memdump", "mm dump")
        self.parser = self.get_argparser()

    def parse_args(self, arg):
        try:
            return self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

    def find_address(self, addr, heaps: List[mm.MMHeap] = None, log=None):
        """Find the node that contains the address from memdump log or live dump."""
        addr = int(utils.parse_and_eval(addr))
        if log:
            nodes = parse_memdump_log(log)
            node = next((node for node in nodes if node.contains(addr)), None)
        else:
            # Find pool firstly
            node = next(
                (blk for pool in mm.get_pools(heaps) if (blk := pool.find(addr))), None
            )

            # Try heap if not found in pool
            node = node or next(
                (node for heap in heaps if (node := heap.find(addr))), None
            )

        return addr, node

    def collect_nodes(self, heaps: List[mm.MMHeap], log=None, filters=None):
        if log:
            nodes = parse_memdump_log(log, filters=filters)
        else:
            nodes = dump_nodes(heaps, filters)

        return nodes

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        if not (args := self.parse_args(arg)):
            return

        print_header()

        pids = [int(tcb["pid"]) for tcb in utils.get_tcbs()]

        def printnode(node, count):
            print_node(node, node.pid in pids, count, no_backtrace=args.no_backtrace)

        heaps = get_heaps(args.heap)

        # Find the node by address, find directly and then quit
        if args.address:
            addr, node = self.find_address(args.address, heaps, args.log)
            if not node:
                print(f"Address {addr:#x} not found in any heap")
            else:
                source = "Pool" if node.from_pool else "Heap"
                printnode(node, 1)
                state = (
                    "unknown"
                    if node.is_free is None
                    else "free" if node.is_free else "used"
                )
                print(f"{addr: #x} is {state}, found belongs to {source} - {node}")

                if node.prevnode:
                    print(f"prevnode: {node.prevnode}")
                if node.nextnode:
                    print(f"nextnode: {node.nextnode}")
            return

        filters = {
            "pid": args.pid,
            "nodesize": args.size,
            "used": not args.free,
            "free": args.free,
            "seqmin": args.min,
            "seqmax": args.max,
            "orphan": args.orphan,
            "no_heap": args.no_heap,
            "no_pool": args.no_pool,
            "no_pid": args.no_pid,
        }

        nodes = self.collect_nodes(heaps, log=args.log, filters=filters)

        sort_method = {
            "count": lambda node: 1,
            "size": lambda node: node.nodesize,
            "nodesize": lambda node: node.nodesize,
            "seq": lambda node: node.seqno,
            "address": lambda node: node.address,
        }

        def sort_nodes(nodes, sort=None):
            sort = sort or args.sort
            nodes = sorted(nodes, key=sort_method[sort], reverse=not args.no_reverse)
            if args.top is not None:
                nodes = nodes[: args.top] if args.top > 0 else nodes[args.top :]
            return nodes

        if args.biggest:
            # Dump the biggest node is same as sort by nodesize and do not group them
            args.sort = "nodesize"
            args.no_group = True

        if args.no_group:
            # Print nodes without grouping
            nodes = list(nodes)

            for node in sort_nodes(nodes):
                printnode(node, 1)

            gdb.write(f"Total blks: {len(nodes)}\n")
        else:
            # Group the nodes and then print

            grouped: Dict[MMNodeDump, MMNodeDump] = defaultdict(list)
            grouped = group_nodes(nodes)

            # Replace the count and size to count grouped nodes
            sort_method["count"] = lambda node: len(grouped[node])
            sort_method["size"] = lambda node: node.nodesize * len(grouped[node])
            total_blk = total_size = 0
            for node in sort_nodes(grouped.keys()):
                count = len(grouped[node])
                total_blk += count
                if node.pid != mm.PID_MM_MEMPOOL:
                    total_size += count * node.nodesize
                printnode(node, count)

            print(f"Total {total_blk} blks, {total_size} bytes")


@autocompeletion.complete
class MMfrag(gdb.Command):
    """Show memory fragmentation rate and analyze fragmentation causes"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "--heap",
            type=str,
            default=None,
            help="Which heap to inspect",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="Show detailed fragmentation causes (orphan nodes between free nodes)",
        )
        parser.add_argument(
            "--threshold",
            type=int,
            default=512,
            help="Minimum total size of adjacent free nodes to consider (bytes, default: 512)",
        )
        parser.add_argument(
            "--top", type=int, default=None, help="Number of top fragments to show"
        )
        parser.add_argument(
            "--no-backtrace",
            action="store_true",
            help="Do not print backtrace",
        )
        return parser

    def __init__(self):
        super().__init__("mm frag", gdb.COMMAND_USER)
        utils.alias("memfrag", "mm frag")
        self.parser = self.get_argparser()

    def show_fragmentation_rate(self, heap: mm.MMHeap):
        """Calculate and display fragmentation rate for a heap"""
        free_nodes = list(
            sorted(heap.nodes_free(), key=lambda node: node.nodesize, reverse=True)
        )
        if not free_nodes:
            gdb.write(f"{heap}: no free nodes\n")
            return

        total_free_size = sum(node.nodesize for node in free_nodes)
        remaining = total_free_size
        frag_rate = 0.0

        for node in free_nodes:
            frag_rate += (1 - (node.nodesize / remaining)) * (
                node.nodesize / total_free_size
            )
            remaining -= node.nodesize

        frag_rate *= 1000
        gdb.write(
            f"{heap.name}@{heap.address:#x}, fragmentation rate:{frag_rate:.2f},"
            f" heapsize: {heap.heapsize}, free size: {total_free_size},"
            f" free count: {len(free_nodes)}, largest: {free_nodes[0].nodesize}\n"
        )

    def show_memory_fragments(self, heap: mm.MMHeap, args):
        """Analyze and display fragmentation causes (orphan nodes between free nodes)"""
        formatter = "{:>4} {:>8} {:>12} {:>12} {:>12} {:>14} {:>9} {:>18} {:}\n"

        def print_fragment_header():
            head = (
                "Pool",
                "PID",
                "PrevSize",
                "NodeSize",
                "NextSize",
                "TotalFreeSize",
                "Seqno",
                "Address",
                "Backtrace",
            )
            gdb.write(formatter.format(*head))

        def print_fragment_node(node: MMNodeDump, total_free, no_backtrace=False):
            gdb.write(
                formatter.format(
                    "P" if node.from_pool else "H",
                    node.pid,
                    node.prevnode.nodesize,
                    node.nodesize,
                    node.nextnode.nodesize,
                    total_free,
                    node.seqno,
                    hex(node.address),
                    "",
                )
            )
            if (
                mm.MM_RECORD_STACK_DEPTH > 0
                and not no_backtrace
                and node.backtrace
                and node.backtrace[0]
            ):
                leading = formatter.format("", "", "", "", "", "", "", "", "")[:-1]
                bt_format = leading + "{1:<48}{2}\n"
                gdb.write(
                    f"{backtrace.Backtrace(node.backtrace, formatter=bt_format)}\n"
                )

        def collect_significant_fragments():
            heap_guards = [item for start, end in heap.regions for item in (start, end)]

            for node in heap.nodes_used():
                if node in heap_guards:
                    continue
                if not (node.is_prev_free and node.nextnode and node.nextnode.is_free):
                    continue
                total_free = sum(n.nodesize for n in [node.prevnode, node.nextnode])
                if total_free < args.threshold:
                    continue

                yield node, total_free

        fragments = list(collect_significant_fragments())
        if not fragments:
            gdb.write(
                f"No significant memory fragments found in {heap.name} "
                f"(threshold: {args.threshold} bytes)\n"
            )
            return

        gdb.write("-" * 106 + "\n")
        print_fragment_header()

        fragments.sort(key=lambda x: x[1], reverse=True)
        for node, total_free in fragments[: args.top]:
            print_fragment_node(node, total_free, args.no_backtrace)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        parsed_args = self.parser.parse_args(gdb.string_to_argv(args))
        if not parsed_args:
            return

        heaps = get_heaps(parsed_args.heap)
        for heap in heaps:
            self.show_fragmentation_rate(heap)
            if parsed_args.verbose:
                self.show_memory_fragments(heap, parsed_args)

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Memory Fragments Report",
            "summary": "Display fragmentation causes (orphan nodes between free nodes)",
            "command": "mm frag",
            "result": "info",
            "category": utils.DiagnoseCategory.memory,
            "message": gdb.execute("mm frag --verbose", to_string=True),
        }


@autocompeletion.complete
class MMMap(gdb.Command):
    """Generate memory map image to visualize memory layout"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-o",
            "--output",
            type=str,
            metavar="file",
            default=None,
            help="img output file",
        )
        parser.add_argument(
            "--heap",
            type=str,
            metavar="symbol",
            help="Which heap's pool to show",
            default=None,
        )
        return parser

    def __init__(self):
        self.np = utils.import_check("numpy", errmsg="Please pip install numpy\n")
        self.plt = utils.import_check(
            "matplotlib", "pyplot", errmsg="Please pip install matplotlib\n"
        )
        self.math = utils.import_check("math")
        if not self.np or not self.plt or not self.math:
            return

        super().__init__("mm map", gdb.COMMAND_USER)
        utils.alias("memmap", "mm map")
        self.parser = self.get_argparser()

    def save_memory_map(self, nodes: List[MMNodeDump], output_file):
        mallinfo = sorted(nodes, key=lambda node: node.address)
        start = mallinfo[0].address
        size = mallinfo[-1].address - start
        order = self.math.ceil(size**0.5)
        img = self.np.zeros([order, order])

        for node in mallinfo:
            addr = node.address
            size = node.nodesize
            start_index = addr - start
            end_index = start_index + size
            img.flat[start_index:end_index] = 1 + self.math.log2(node.seqno + 1)

        self.plt.imsave(output_file, img, cmap=self.plt.get_cmap("Greens"))

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        if not (args := self.parse_arguments(gdb.string_to_argv(args))):
            return

        for heap in get_heaps(args.heap):
            name = heap.name or f"heap@{heap.address:#x}"
            output = args.output or f"{name}.png"
            self.save_memory_map(heap.nodes_used(), output)
            gdb.write(f"Memory map saved to {output}\n")


@autocompeletion.complete
class MMVisualize(gdb.Command):
    """Generates a memory treemap, showing all backtrace statistics"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-o",
            "--output",
            type=str,
            metavar="file",
            default="mm_visualize",
            help="html output file",
        )
        parser.add_argument(
            "--heap",
            type=str,
            metavar="symbol",
            help="Which heap's pool to show",
            default=None,
        )
        return parser

    def __init__(self):
        self.backtrace_depth = mm.MM_RECORD_STACK_DEPTH
        if self.backtrace_depth <= 0:
            gdb.write(
                "Without mm record backtrace enabled, visualization is not possible\n"
            )
            return

        super().__init__("mm visualize", gdb.COMMAND_USER)
        utils.alias("memvisualize", "mm visualize")
        self.px = utils.import_check(
            "plotly.express", errmsg="Please pip install plotly\n"
        )
        self.pd = utils.import_check("pandas", errmsg="Please pip install pandas\n")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        nodes = mm.get_nodes_dict()
        df = self.pd.DataFrame.from_records(nodes)

        # Only show used nodes, exclude all free nodes
        df = df[~df["free"]]

        # Merge all nodes with the same pid and backtrace
        df = (
            df.groupby(["backtrace", "pid"], as_index=False)
            .agg(
                {
                    "name": lambda x: x.iloc[0],
                    "address": "any",
                    "size": "sum",
                    "pid": "first",
                    "from_pool": "first",
                }
            )
            .reset_index()
        )

        for i in range(self.backtrace_depth):
            df[f"stack_{i}"] = df["backtrace"].apply(
                lambda x: (
                    "Unkown" if i >= len(x) or not x[i] else utils.Symbol(x[i]).func
                )
            )

        # Drop the backtrace column
        df = df.drop(columns=["backtrace"])

        # Reorder the backtrace to ensure it is displayed in the correct format in the UI.
        # For example, a,b,c,0,U,U,U,U generates U,U,U,U,U,a,b,c
        stack_columns = [f"stack_{i}" for i in range(8)]
        for index, row in df.iterrows():
            stack_values = row[stack_columns].tolist()
            unknowns = [val for val in stack_values if val == "Unkown"]
            non_unknowns = [val for val in stack_values if val != "Unkown"]
            new_stack_values = unknowns + non_unknowns
            df.loc[index, stack_columns] = new_stack_values

        # Generate the treemap
        stack_cols = [f"stack_{i}" for i in range(self.backtrace_depth - 1, -1, -1)]
        fig = self.px.treemap(
            df,
            path=stack_cols,
            values="size",
            hover_data=["pid", "from_pool"],
            title="Memory Allocation",
        )

        fig.update_layout(
            margin=dict(t=50, l=25, r=25, b=25),
            coloraxis_colorbar=dict(title="Size (bytes)"),
        )
        fig.update_traces(
            maxdepth=3,
            texttemplate="%{label}<br>%{value:,d} bytes",
            textposition="middle center",
            marker=dict(line=dict(width=1, color="DarkGray")),
        )

        path = args.output + ".html"
        fig.write_html(path)
        gdb.write(f"Memory map saved to Memory visualizations saved to {path}\n")


@autocompeletion.complete
class MMFree(gdb.Command):
    """Show heap statistics, same as device command free"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)

        parser.add_argument(
            "--heap",
            type=str,
            default=None,
            metavar="symbol",
            help="Which heap to inspect",
        )

        parser.add_argument(
            "-t",
            "--thread-usage",
            action="store_true",
            help="Show memory usage of each thread",
        )
        return parser

    def __init__(self):
        super().__init__("mm free", gdb.COMMAND_USER)
        utils.alias("free", "mm free")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        heaps = get_heaps(args.heap)

        formatter = "{:<20} {:<10} {:<10} {:<10} {:<10} {:<10} {:<10} {:<10}"
        header = (
            "name",
            "total",
            "used",
            "free",
            "maxused",
            "maxfree",
            "nused",
            "nfree",
        )

        print(formatter.format(*header))
        mm_heap_s = utils.lookup_type("struct mm_heap_s")

        # Print summary of memory usage
        summary = defaultdict(lambda: {"size": 0, "count": 0})
        mempool_free = 0

        for heap in heaps:
            heap_free = heap_used = 0
            total_size = max_free = nused = nfree = 0

            # Heap nodes
            for node in heap.nodes:
                nodesize = node.nodesize
                total_size += nodesize

                if node.is_free:
                    nfree += 1
                    heap_free += nodesize
                    max_free = max(max_free, nodesize)
                else:
                    heap_used += nodesize
                    nused += 1
                    summary[node.pid]["size"] += node.nodesize
                    summary[node.pid]["count"] += 1

            mempool_total = 0
            mempool_maxused = 0
            mempool_used = 0
            mempool_nused = 0
            mempool_nfree = 0
            mempool_free = 0

            if mm_pool := heap.mm_mpool:
                mpool = mm.MemPoolMultiple(mm_pool)

                for chunk in mpool.chunks:
                    mempool_total += int(chunk.end) - int(chunk.start)

                for pool in mpool.pools:
                    mempool_maxused += pool.total
                    for blk in pool.blks:
                        if blk.is_free:
                            mempool_free += blk.nodesize
                            mempool_nfree += 1
                        else:
                            mempool_used += blk.nodesize
                            mempool_nused += 1
                            summary[blk.pid]["size"] += node.nodesize
                            summary[blk.pid]["count"] += 1

            total = heap.heapsize + mm_heap_s.sizeof
            heap_used += mm_heap_s.sizeof  # struct overhead

            print(
                formatter.format(
                    heap.name,
                    total,
                    heap_used - mempool_free,
                    heap_free + mempool_free,
                    int(heap.mm_maxused),
                    max_free,
                    nused,
                    nfree,
                )
            )

            if mempool_total:
                print(
                    formatter.format(
                        f"{heap.name}@mempool",
                        mempool_total,
                        mempool_used,
                        mempool_total - mempool_used,
                        mempool_maxused,
                        mempool_total - mempool_maxused,
                        mempool_nused,
                        mempool_nfree,
                    )
                )

        if args.thread_usage:
            print("\nSummary of thread memory usage:")
            formatter = "{:<10} {:<30} {:<16} {:<10} {:<10}"
            print(formatter.format("PID", "TaskName", "Size", "Size(kB)", "Count"))
            for pid, data in sorted(
                summary.items(), key=lambda x: x[1]["size"], reverse=True
            ):
                print(
                    formatter.format(
                        pid,
                        utils.get_task_name(pid) or "<noname>",
                        data["size"],
                        f"{data['size'] / 1024:.1f}kB",
                        data["count"],
                    )
                )

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Memory Free Information",
            "summary": "Memory free information",
            "command": "free",
            "result": "info",
            "category": utils.DiagnoseCategory.memory,
            "message": gdb.execute("mm free --thread-usage", to_string=True),
        }


@autocompeletion.complete
class NxMemoryRange(gdb.Command):
    """Show RAM range of heap and sections"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument("--heap-only", action="store_true", help="Heap only")
        parser.add_argument(
            "--globals-only", action="store_true", help="Global variables only"
        )
        return parser

    def __init__(self):
        super().__init__("mm range", gdb.COMMAND_USER)
        utils.alias("memrange", "mm range")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        memrange = mm.get_memrange(None, args.heap_only, args.globals_only)
        if not memrange:
            print("No memory range found")
            return

        header = ("start", "end", "size", "size(kB)")
        formatter = "{:<20} {:<20} {:<20} {:<20}"
        print(formatter.format(*header))
        for start, end in memrange:
            length = end - start
            print(
                formatter.format(hex(start), hex(end), length, f"{length / 1024: .1f}")
            )


@autocompeletion.complete
class NxDumpRAM(gdb.Command):
    """Dump memory to file, similar to GDB dump memory"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-o",
            "--output",
            help="Memory dump output directory",
            metavar="file",
            default="memdump",
        )
        parser.add_argument("-r", "--memrange", type=str, default=None)
        parser.add_argument("--heap-only", action="store_true", help="Heap only")
        parser.add_argument(
            "--globals-only", action="store_true", help="Global variables only"
        )
        return parser

    def __init__(self):
        super().__init__("dump ram", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        memrange = mm.get_memrange(args.memrange, args.heap_only, args.globals_only)
        if not memrange:
            print("No memory range found")
            return

        # Enable trust-readonly flag to try to read from elf file if possible
        gdb.execute("set trust-readonly-sections on")

        print(f"Dumping memory to {args.output}")
        os.makedirs(args.output, exist_ok=True)

        for start, end in memrange:
            print(f"Dumping memory range {start:#x} - {end:#x}")
            try:
                data = gdb.selected_inferior().read_memory(start, end - start)
                output = os.path.join(args.output, f"{start:#x}.bin")
            except gdb.MemoryError:
                print(f"Failed to read memory range {start:#x} - {end:#x}")
                continue

            with open(output, "wb") as f:
                f.write(data)


@autocompeletion.complete
class NxMemoryFind(gdb.Command):
    """Find memory address by pattern"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument("pattern", type=str, help="Pattern to search")
        parser.add_argument("-r", "--memrange", type=str, default=None)
        parser.add_argument("--heap-only", action="store_true", help="Heap only")
        parser.add_argument(
            "--globals-only", action="store_true", help="Global variables only"
        )
        return parser

    def __init__(self):
        super().__init__("mm find", gdb.COMMAND_USER)
        utils.alias("memfind", "mm find")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        if args.pattern.startswith('"') and args.pattern.endswith('"'):
            # Search for string
            value = bytes(args.pattern[1:-1], "utf-8")
        else:
            # Convert to a number
            value = utils.parse_arg(args.pattern)
            value = value.to_bytes((value.bit_length() + 7) // 8, "little")

        print(f"Searching for pattern {binascii.hexlify(value)} in memory")

        memrange = mm.get_memrange(args.memrange, args.heap_only, args.globals_only)
        for start, end in memrange:
            try:
                data = bytes(gdb.selected_inferior().read_memory(start, end - start))
            except gdb.MemoryError:
                print(f"Failed to read memory range {start:#x} - {end:#x}")
                continue

            # Find all occurrences of the byte pattern
            offset = 0
            while True:
                offset = data.find(value, offset)
                if offset == -1:
                    break
                print(f"Found pattern @ {offset + start:#x}")
                offset += 1
        print("Done")



#! /usr/bin/env python3
############################################################################
# tools/pynuttx/nxelf/elf.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################


import bisect
import functools
import hashlib
import importlib
import logging
import os
import pickle
import sys
import time
from enum import IntEnum
from typing import Tuple

from .macros import Macro

try:
    import cxxfilt
    import lief
    from construct import (
        Array,
        Construct,
        Float16l,
        Float32l,
        Float64l,
        Int8sl,
        Int8ul,
        Int16sl,
        Int16ul,
        Int32sb,
        Int32sl,
        Int32ub,
        Int32ul,
        Int64sb,
        Int64sl,
        Int64ub,
        Int64ul,
        Struct,
    )
    from elftools.elf.elffile import ELFFile
    from elftools.elf.sections import SymbolTableSection
except ModuleNotFoundError as e:
    print(f"Error:{e}.\nPlease execute the following command to install dependencies:")
    print("pip install construct pyelftools cxxfilt lief ")

log = logging.getLogger(__name__)


class TypeConflictError(Exception):
    """
    Symbols have different definitions
    """

    pass


class Types:
    def __init__(self, tag):
        self.types = {}
        self.tag = tag
        self.result = dict()

    def set_type(self, die):
        if "DW_AT_name" not in die.attributes:
            return

        name = die.attributes["DW_AT_name"].value.decode("utf-8")
        if name not in self.types:
            self.types[name] = set()

        self.types[name].add(die)

    def get_types(self, type_name):
        if type_name in self.types:
            sets = self.types[type_name]
            return sets
        else:
            return None

    def set_result(self, type_name, result):
        if len(result) != 1:
            raise TypeConflictError(
                f"Multiple different definitions or values ​​exist for a symbol: {type_name} {result}"
            )

        result = result.pop()
        if type_name in self.types:
            self.result[type_name] = result
        return result

    def get_result(self, type_name):
        if type_name in self.result:
            return self.result[type_name]
        return None


class ELFParser:
    """
    ELF file parser class for extracting the following information from ELF files:
    - Symbol addresses
    - Structure definitions
    - Enumeration type definitions
    - Enumeration values

    Main functionality:
    1. Get structure definitions
        elf_parser = ELFParser("nuttx")
        struct = elf_parser.get_type("file_operations")  # Returns construct.Struct object
        print(struct._subcons)

        # result:
        Container:
            open = <Renamed open <FormatField>>
            close = <Renamed close <FormatField>>
            read = <Renamed read <FormatField>>
            write = <Renamed write <FormatField>>
            seek = <Renamed seek <FormatField>>
            ioctl = <Renamed ioctl <FormatField>>
            mmap = <Renamed mmap <FormatField>>
            truncate = <Renamed truncate <FormatField>>
            poll = <Renamed poll <FormatField>>
            unlink = <Renamed unlink <FormatField>>

    2. Get symbol addresses
       addr = elf_parser.symbol_addr("_SeggerRTT")  # Returns symbol address

    3. Get enumeration type definitions
       enum = elf_parser.get_type("tstate_e")  # Returns construct.IntEnum object

    4. Get enumeration values
        value = elf_parser.enum_value("TSTATE_TASK_RUNNING")  # Returns integer value

    """

    def __init__(self, elf_path):
        self.elf = ELFFile(open(elf_path, "rb"))
        self.addr = list()
        self.types = dict()
        self.info = dict()
        self.symbol = dict()
        self.result = dict()
        self.dwarf = self.elf.get_dwarf_info()

        t = time.time()
        print("Parsing ELF file...")
        self.parse_header()
        self.parse_symbol()
        self.macro = Macro(elf_path)
        print(f"ELF file parsed in {time.time() - t:.1f} seconds")

    def parse_header(self):
        header = self.elf.header
        self.info["bitwides"] = (
            32 if header["e_ident"]["EI_CLASS"] == "ELFCLASS32" else 64
        )
        self.info["byteorder"] = (
            "little" if header["e_ident"]["EI_DATA"] == "ELFDATA2LSB" else "big"
        )
        self.info["arch"] = header["e_machine"]
        self.info["size_t"] = "uint%d" % self.info["bitwides"]

    def symbol_filter(self, symbol):
        if symbol["st_info"]["type"] != "STT_FUNC":
            return None
        if symbol["st_info"]["bind"] == "STB_WEAK":
            return None
        if symbol["st_shndx"] == "SHN_UNDEF":
            return None
        return symbol

    def parse_symbol(self):
        tables = [
            s
            for _, s in enumerate(self.elf.iter_sections())
            if isinstance(s, SymbolTableSection)
            and s.name == ".symtab"
            and s["sh_entsize"]
        ]

        for section in tables:
            for nsym, symbol in enumerate(section.iter_symbols()):
                try:
                    name = cxxfilt.demangle(symbol.name)
                except Exception:
                    name = symbol.name
                self.symbol[name] = symbol["st_value"]
                if symbol_filter := self.symbol_filter(symbol):
                    self.addr.append(
                        (symbol_filter["st_value"], symbol_filter["st_size"], name)
                    )

        self.addr.sort(key=lambda x: x[0])

    @functools.lru_cache(maxsize=None)
    def symbol_addr(self, name):
        if len(self.symbol.keys()) == 0:
            self.parse_symbol()

        if name not in self.symbol:
            return None

        return self.symbol[name]

    @functools.lru_cache(maxsize=None)
    def addr2symbol(self, address):
        """Get function name from address"""
        if len(self.addr) == 0:
            self.parse_symbol()

        if i := bisect.bisect_right(self.addr, address, key=lambda x: x[0]):
            addr, size, name = self.addr[i - 1]
            if addr <= address < addr + size:
                return name

        return None

    @functools.lru_cache(maxsize=None)
    def readstring(self, addr):
        for segment in self.elf.iter_segments():
            seg_addr = segment["p_paddr"]
            seg_size = min(segment["p_memsz"], segment["p_filesz"])
            if addr >= seg_addr and addr <= seg_addr + seg_size:
                data = segment.data()[addr - seg_addr :]
                data = data.split(b"\x00")[0]
                return data.decode("utf-8")
        return None

    @functools.lru_cache(maxsize=None)
    def read(self, addr, size):
        for segment in self.elf.iter_segments():
            seg_addr = segment["p_paddr"]
            seg_size = min(segment["p_memsz"], segment["p_filesz"])
            if addr >= seg_addr and addr + size <= seg_addr + seg_size:
                data = segment.data()
                start = addr - seg_addr
                return data[start : start + size]

    @functools.lru_cache(maxsize=None)
    def parse_base_type(self, die):
        name = die.attributes["DW_AT_name"].value.decode("utf-8")
        size = die.attributes["DW_AT_byte_size"].value

        basetypes = {
            "unsigned": {1: Int8ul, 2: Int16ul, 4: Int32ul, 8: Int64ul},
            "float": {2: Float16l, 4: Float32l, 8: Float64l},
            "signed": {1: Int8sl, 2: Int16sl, 4: Int32sl, 8: Int64sl},
        }

        if "unsigned" in name:
            return basetypes["unsigned"].get(size)
        elif "double" in name or "float" in name:
            return basetypes["float"].get(size)
        elif "_Bool" in name:
            return Int8ul
        elif "char" in name or "short" in name or "int" in name:
            return basetypes["signed"].get(size)

        raise ValueError(f"Unsupported base type: {name}")

    @functools.lru_cache(maxsize=None)
    def parse_array(self, die):
        nums = 0
        for child in die.iter_children():
            if "DW_AT_upper_bound" in child.attributes:
                nums = child.attributes["DW_AT_upper_bound"].value + 1
            elif "DW_AT_count" in child.attributes:
                nums = child.attributes["DW_AT_count"].value

        type_die = self.dwarf.get_DIE_from_refaddr(
            die.attributes["DW_AT_type"].value + die.cu.cu_offset
        )

        item_type = self.parse_die(type_die)

        def dynamic_array(ctx):
            # If array_field is specified during parse, array parsing uses the specified field as the length
            if hasattr(ctx, "_params") and "array_field" in ctx._params:
                field_name = ctx._params["array_field"]
                if hasattr(ctx, field_name):
                    return getattr(ctx, field_name)
            elif hasattr(ctx, "_params") and "array_length" in ctx._params:
                # If array_length is specified during parse, the array is parsed using the specified length.
                return ctx._params["array_length"]

            return nums

        """
            Usage example:
            struct example_s {
                int length;
                int buffer[0];
            };
            struct = elf.get_type("example")
            struct.parse(
                b"\x00\x00\x00\x04"  # length = 4
                b"\x01\x02\x03\x04",  # buffer = [1, 2, 3, 4]
                array_field="length"  # Use the length field to determine the size of the array
            )
        """

        return Array(dynamic_array, item_type)

    @functools.lru_cache(maxsize=None)
    def parse_typedef(self, die):
        type_attr = die.attributes["DW_AT_type"]
        die = self.dwarf.get_DIE_from_refaddr(type_attr.value + die.cu.cu_offset)
        return self.parse_die(die)

    @functools.lru_cache(maxsize=None)
    def parse_struct(self, die):
        members = dict()
        for child in die.iter_children():
            member_name = child.attributes["DW_AT_name"].value.decode("utf-8")
            member_type = child.attributes["DW_AT_type"].value
            type_die = self.dwarf.get_DIE_from_refaddr(member_type + die.cu.cu_offset)
            member_type = self.parse_die(type_die)
            members[member_name] = member_type

        struct = Struct(**members)
        return struct

    @functools.lru_cache(maxsize=None)
    def parse_enum(self, die) -> IntEnum:
        if die.tag != "DW_TAG_enumeration_type":
            raise ValueError(f"type is not enum: {die.tag}")

        enum = dict()
        name = die.attributes["DW_AT_name"].value.decode("utf-8")
        for child in die.iter_children():
            name = child.attributes["DW_AT_name"].value.decode("utf-8")
            value = child.attributes["DW_AT_const_value"].value
            enum[name] = value

        return IntEnum(name, enum)

    @functools.lru_cache(maxsize=None)
    def parse_enum_value(self, die):
        if die.tag != "DW_TAG_enumerator":
            raise ValueError(f"type is not enum: {die.tag}")

        name = die.attributes["DW_AT_name"].value.decode("utf-8")
        value = die.attributes["DW_AT_const_value"].value
        return name, value

    @functools.lru_cache(maxsize=None)
    def find_die_by_name(self, name):
        if name in self.types:
            return self.types[name]

        for CU in self.dwarf.iter_CUs():
            result = None
            for DIE in CU.iter_DIEs():
                if "DW_AT_name" not in DIE.attributes:
                    continue

                AT_name = DIE.attributes["DW_AT_name"].value.decode("utf-8")
                self.types[AT_name] = DIE
                if name == AT_name:
                    result = DIE

            if result:
                return result

        return None

    @functools.lru_cache(maxsize=None)
    def parse_die(self, die):
        if (
            "DW_AT_name" not in die.attributes
            and die.tag != "DW_TAG_pointer_type"
            and die.tag != "DW_TAG_array_type"
        ):
            return None

        tag_handlers = {
            "DW_TAG_structure_type": self.parse_struct,
            "DW_TAG_enumeration_type": self.parse_enum,
            "DW_TAG_base_type": self.parse_base_type,
            "DW_TAG_typedef": self.parse_typedef,
            "DW_TAG_pointer_type": lambda _: (
                Int32ul if self.info["bitwides"] == 32 else Int64ul
            ),
            "DW_TAG_array_type": self.parse_array,
        }

        if die.tag not in tag_handlers:
            raise ValueError(f"Unsupported type: {die.tag}")

        return tag_handlers[die.tag](die)

    @functools.lru_cache(maxsize=None)
    def get_type(self, type_name):
        if type_name in self.result:
            return self.result[type_name]

        die = self.find_die_by_name(type_name)
        if die is None:
            return None

        return self.parse_die(die)


class LiefELF:
    def __init__(self, filename):
        self.elf = lief.parse(filename)
        if not self.elf:
            raise BaseException(f"Failed to parse ELF file: {filename}")

        self.endian = (
            "l"
            if self.elf.abstract.header.endianness == lief.Header.ENDIANNESS.LITTLE
            else "b"
        )
        self.architecture = self.elf.abstract.header.architecture
        self.bits = 64 if self.elf.abstract.header.is_64 else 32

    def get_symbol(self, symbol):
        # Try to get LTO private symbol if not found.
        # This may still not work if LTO renames the symbol, or
        # it has different suffix.
        # Try to get symbol with .0 suffix too for some static variable defined in function scope, like g_statenames
        return (
            self.elf.get_symbol(symbol)
            or self.elf.get_symbol(f"{symbol}.lto_priv.0")
            or self.elf.get_symbol(f"{symbol}.0")
        )

    def read_symbol(
        self, symbol, struct: Construct = None
    ) -> Tuple[lief.Symbol, memoryview]:
        sym = self.get_symbol(symbol)
        if sym is None:
            return None

        data = self.read_from(sym.value, sym.size)
        if struct:
            data = struct.parse(data)
        return sym, data

    def read_from(self, addr, len=1) -> memoryview:
        for section in self.elf.sections:
            if section.type == lief.ELF.Section.TYPE.PROGBITS:
                off = addr - section.virtual_address
                if (
                    section.virtual_address
                    <= addr
                    < section.virtual_address + section.size
                    and section.size - off >= len
                ):
                    return section.content[off : off + len]

        for segment in self.elf.segments:
            if segment.type == lief.ELF.Segment.TYPE.LOAD:
                off = addr - segment.virtual_address
                if (
                    segment.virtual_address
                    <= addr
                    < segment.virtual_address + segment.virtual_size
                    and segment.virtual_size - off >= len
                ):
                    return segment.content[off : off + len]

        return None

    def read_string(self, addr) -> str:
        """Read const string from ELF file"""
        output = b""
        while True:
            c = self.read_from(addr, 1)
            if c == b"\0":
                break

            output += c.tobytes()
            addr += 1

        return output.decode("utf-8", errors="replace")

    def get_inttype(self) -> Construct:
        return {
            "32l": Int32sl,
            "32b": Int32sb,
            "64l": Int64sl,
            "64b": Int64sb,
        }.get(f"{self.bits}{self.endian}", Int32sl)

    def get_pointer_type(self) -> Construct:
        return {
            "32l": Int32ul,
            "32b": Int32ub,
            "64l": Int64ul,
            "64b": Int64ub,
        }.get(f"{self.bits}{self.endian}", Int32ul)

    def get_pointer_size(self):
        return 8 if self.elf.abstract.header.is_64 else 4


class AngrElf:
    def __init__(self, elf: str):
        try:
            self.angr = importlib.import_module("angr")
            self.capstone = importlib.import_module("capstone")
        except Exception as e:
            print(
                f"Error:{e}.\nPlease execute the following command to install dependencies:"
            )
            print("pip install angr capstone")
            sys.exit(1)

        with open(elf, "rb") as f:
            hash = hashlib.md5(f.read()).hexdigest()

        project_cache_path = f"{hash}.project"
        cfg_cache_path = f"{hash}.cfg"

        self.project = self.load_save_cache(
            project_cache_path, lambda: self.angr.Project(elf, auto_load_libs=False)
        )
        self.cfg = self.load_save_cache(
            cfg_cache_path, lambda: self.project.analyses.CFGFast()
        )

        for func in self.cfg.kb.functions.values():
            func._project = self.project

    def load_save_cache(self, file_path, func):
        cache_path = f"{file_path}.pkl"
        if os.path.exists(cache_path):
            with open(cache_path, "rb") as f:
                print(f"load cache {cache_path}")
                return pickle.load(f)
        else:
            print(f"not found cache {cache_path}")
            obj = func()
            with open(cache_path, "wb") as f:
                print(f"create cache {cache_path}")
                pickle.dump(obj, f)
            return obj

    def sym2addr(self, sym: str):
        sym = self.cfg.kb.functions.floor_func(sym)
        if sym:
            return sym.addr
        else:
            return None

    def addr2sym(self, addr: int):
        sym = self.cfg.kb.functions.floor_func(addr)
        if sym:
            return sym.name
        else:
            return f"0x{addr:08x}"

    def addr2func(self, addr: int):
        addr = addr
        sym = self.cfg.kb.functions.floor_func(addr)
        if sym:
            return self.cfg.kb.functions[sym.addr]
        else:
            return None

    def sym2func(self, sym: str):
        addr = self.sym2addr(sym)
        if addr:
            return self.cfg.kb.functions[addr]
        else:
            return None

    def addr2block(self, addr: int):
        addr = addr
        sym = self.cfg.kb.functions.floor_func(addr)
        if sym:
            func = self.cfg.kb.functions[sym.addr]
            for block in func._addr_to_block_node.values():
                if addr >= block.addr and addr < block.addr + block.size:
                    return block
        else:
            return None

    def print_block(self, block):
        print(f"basic block: {hex(block.addr)}, size: {block.size}")
        for insn in block.capstone.insns:
            print(f"    {hex(insn.address)}: {insn.mnemonic} {insn.op_str}")
            if insn.group(self.capstone.CS_GRP_CALL):
                print(f" call {hex(insn.operands[0].imm)}")
            elif insn.group(self.capstone.CS_GRP_RET):
                print(f" return {hex(insn.operands[0].imm)}")
            elif insn.group(self.capstone.CS_GRP_INT):
                print(f" interrupt {hex(insn.operands[0].imm)}")
            elif insn.group(self.capstone.CS_GRP_BRANCH_RELATIVE):
                target_addr = insn.operands[0].imm
                target_function = self.cfg.kb.functions.floor_func(target_addr)
                print(
                    f" conditional jump {hex(target_addr)} (target function: {target_function.name})"
                )
            elif insn.group(self.capstone.CS_GRP_JUMP):
                sub_function = self.cfg.kb.functions.floor_func(insn.operands[0].imm)
                print(
                    f" unconditional jump {hex(insn.operands[0].imm)} (target function: {sub_function.name})"
                )
        sys.stdout.flush()

    def print_function(self, function):
        print(f"function: {function.name} {function.addr} {function.size}")
        for block in function.blocks:
            self.print_block(block)

    def get_block_next_address(self, block):
        for insn in block.capstone.insns:
            if insn.group(self.capstone.CS_GRP_BRANCH_RELATIVE):
                try:
                    addr = insn.operands[0].imm
                    func = self.addr2func(addr)
                    if func:
                        next_block = func.get_block(addr)
                        return next_block.addr
                except Exception as e:
                    log.error(
                        f"get_block_next_address failed: addr: {block.addr:#08x} {e}"
                    )
                    return None
        return None



############################################################################
# tools/pynuttx/nxgdb/libuv.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb

from . import autocompeletion, lists, utils

CONFIG_LIBUV = utils.lookup_type("struct uv_loop_s") is not None


@autocompeletion.complete
class UVDump(gdb.Command):
    """Dump libuv loop and handle informations"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-l",
            "--loop",
            type=str,
            required=True,
            help="Memory address of uv_loop structure",
        )
        parser.add_argument(
            "--handles",
            action="store_true",
            help="Dump handles information instead of loop data",
        )
        parser.add_argument(
            "--active",
            action="store_true",
            help="Filter only active handles (requires --handles)",
        )
        parser.add_argument(
            "--no-backtrace",
            action="store_true",
            help="Do not print backtrace (requires --handles)",
        )
        return parser

    def __init__(self):
        if not CONFIG_LIBUV:
            return
        super(UVDump, self).__init__("uvdump", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def dump_loop_info(self, loop_ptr):
        """Dump uv_loop_t structure fields"""
        if not loop_ptr:
            return

        print(loop_ptr.dereference())

    def dump_uv_handles(self, loop_ptr, active_only=False, no_backtrace=False):
        """Dump handle information from uv_loop_t structure"""
        if not loop_ptr:
            return

        UV_HANDLE_FLAG = utils.enum("enum uv_handle_flag")

        formatter = "{:>8} {:>12} {:>18} {:}\n"
        headers = ("Flags", "Type", "Address", "Backtrace")
        gdb.write(formatter.format(*headers))
        gdb.write("-" * 50 + "\n")

        try:
            handle_queue_head = loop_ptr["handle_queue"].address
            handle_queue = lists.NxList(
                handle_queue_head, utils.lookup_type("uv_handle_t"), "handle_queue"
            )
            type_name_map = {
                member.value: member.name.lower()
                for member in utils.enum("uv_handle_type")
            }

            for uv_handle in handle_queue:
                handle_flags = int(uv_handle["flags"])
                is_active = handle_flags & UV_HANDLE_FLAG.HANDLE_ACTIVE.value

                if active_only and not is_active:
                    continue

                handle_type = int(uv_handle["type"])
                type_name = type_name_map.get(handle_type, f"<unknown:{handle_type}>")
                flag_str = (
                    "["
                    + ("R" if handle_flags & UV_HANDLE_FLAG.HANDLE_REF.value else "-")
                    + (
                        "A"
                        if handle_flags & UV_HANDLE_FLAG.HANDLE_ACTIVE.value
                        else "-"
                    )
                    + (
                        "I"
                        if handle_flags & UV_HANDLE_FLAG.HANDLE_INTERNAL.value
                        else "-"
                    )
                    + "]"
                )

                gdb.write(
                    formatter.format(
                        flag_str,
                        type_name,
                        f"0x{int(uv_handle):x}",
                        "",
                    )
                )

                if not no_backtrace:
                    backtrace = tuple(utils.BacktraceEntry(uv_handle["stack"]).get())
                    if backtrace and backtrace[0]:
                        leading = formatter.format("", "", "", "")[:-1]
                        bt_format = leading + "{1:<48}{2}\n"
                        gdb.write(f"{utils.Backtrace(backtrace, formatter=bt_format)}")
        except gdb.error:
            print("Error: Unable to access handle queue.")
            return

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            parsed_args = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        if not parsed_args.handles and (parsed_args.active or parsed_args.no_backtrace):
            required = []
            if parsed_args.active:
                required.append("--active")
            if parsed_args.no_backtrace:
                required.append("--no-backtrace")

            print(f"Error: Options {', '.join(required)} require --handles")
            return

        loop = utils.gdb_eval_or_none(parsed_args.loop)
        if not loop:
            return

        loop_ptr = loop.cast(utils.lookup_type("uv_loop_t").pointer())
        if parsed_args.handles:
            self.dump_uv_handles(loop_ptr, parsed_args.active, parsed_args.no_backtrace)
        else:
            self.dump_loop_info(loop_ptr)



############################################################################
# tools/pynuttx/gdbrpc/__init__.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

__all__ = [
    "Client",
    "ClientCLI",
    "PostRequest",
    "Request",
    "Response",
    "PacketStatus",
    "Server",
    "ShellExec",
]

from .cli import ClientCLI

# Client must be imported first because ClientCLI depends on it
from .client import Client
from .utils import PacketStatus, PostRequest, Request, Response, ShellExec

# Register GDB commands if running inside GDB
try:
    from gdb import COMMAND_USER, Command

    from .commands import (  # noqa: F401
        SocketServerStatus,
        StartSocketClient,
        StartSocketServer,
        StopSocketServer,
    )

    class GDBRpcPrefix(Command):
        """GDB Remote Protocol related commands prefix"""

        def __init__(self):
            super().__init__("gdbrpc", COMMAND_USER, prefix=True)

    print("Registering gdbrpc commands...")
    GDBRpcPrefix()
    StartSocketServer()
    StopSocketServer()
    SocketServerStatus()
    StartSocketClient()
except ImportError:
    pass



############################################################################
# tools/pynuttx/nxgdb/nxcrash/stackoverflow.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb

from .. import backtrace, utils
from ..stack import Stack

STACK_FILL_THRESHOLD = 0.9


class CrashStackOverflow(gdb.Command):
    """Analyse and collect the stack overflow threads"""

    def __init__(self):
        super().__init__("crash stackoverflow", gdb.COMMAND_USER)

    def collect(self, tcbs):
        """Collect the stack overflow information"""

        collected = []
        for tcb in tcbs:
            st = Stack(
                int(tcb["stack_base_ptr"]),
                int(tcb["adj_stack_size"]),
                utils.get_sp(tcb),
            )

            filled = st.max_usage() / st._stack_size
            if filled > STACK_FILL_THRESHOLD:
                collected.append((tcb, filled))

        return collected

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        collected = self.collect(utils.get_tcbs())
        if not collected:
            gdb.write("No stack overflow found\n")
            return

        print(
            f"Found stack overflow threads\n{'PID':<4} {'NAME':<10} {'STACKSIZE':<10} {'FILLED'}"
        )
        for tcb, filled in collected:
            print(
                f"{tcb['pid']:<4} {utils.get_task_name(tcb):<10} {tcb.adj_stack_size:<10} {filled:.2%}"
            )

    def diagnose(self, *args, **kwargs):
        collected = self.collect(utils.get_tcbs())
        return {
            "title": "Stack Overflow Report",
            "summary": (
                f"{'No' if not collected else len(collected)} threads{'s' if len(collected) != 1 else ''} found"
            ),
            "result": "fail" if collected else "pass",
            "category": utils.DiagnoseCategory.memory,
            "command": "crash stackoverflow",
            "thread": [
                {
                    "pid": tcb["pid"],
                    "name": utils.get_task_name(tcb),
                    "stacksize": tcb.adj_stack_size,
                    "filled": filled,
                    "backtrace": backtrace.Backtrace(
                        utils.get_backtrace(int(tcb["pid"])), break_null=False
                    ),
                }
                for tcb, filled in collected
            ],
        }



############################################################################
# tools/pynuttx/gdbrpc/server.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import os
import queue
import socket
import sys
import threading
import traceback
from datetime import datetime
from typing import Any, Dict, Optional, Tuple

import cloudpickle as pickle
import gdb
from gdbrpc.utils import (
    DEFAULT_TIMEOUT,
    PacketStatus,
    Request,
    Response,
    socket_recv,
    socket_send,
)


# Custom thread class for GDB compatibility
# In newer GDB versions (>= 14.1), gdb.Thread is available as a subclass of threading.Thread
# that automatically blocks signals. For older versions, we implement it ourselves.
class GdbThread(threading.Thread):
    """
    A thread class that works seamlessly with GDB's signal handling requirements.
    This is a drop-in replacement for gdb.Thread in newer GDB versions.
    """

    def start(self):
        """Override start to block signals before starting the thread."""
        # Check if gdb.blocked_signals is available (GDB >= 13.1)
        if hasattr(gdb, "blocked_signals"):
            with gdb.blocked_signals():
                super().start()
        else:
            # Fall back to regular threading for older GDB versions
            super().start()


class AsyncExec:
    def __init__(self, request: Request):
        self.request: Request = request
        self._queue = queue.Queue()

    def __call__(self):
        self.request(self._queue)

    def get_result(self, timeout: float = DEFAULT_TIMEOUT) -> Any:
        try:
            return self._queue.get(timeout=timeout)
        except queue.Empty:
            raise TimeoutError("No result available within the specified timeout")


class Server:
    def __init__(
        self,
        host: str = "localhost",
        port: int = 20819,
        log_level: int = logging.INFO,
        log_path: Optional[str] = None,
    ):
        self.port = port
        self.host = host
        self.server: socket.socket
        self.running = False
        self.accept_thread: Optional[GdbThread] = None
        self.clients_lock = threading.Lock()
        self.clients: Dict[Tuple[str, int], socket.socket] = {}

        self._logger = logging.getLogger(__name__)
        if not self._logger.hasHandlers():
            self._logger.setLevel(log_level)

            formatter = logging.Formatter("%(asctime)s gdbrpc_server: %(message)s")

            if log_path is None:
                timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
                pid = os.getpid()
                log_path = f"gdbrpc_server-{timestamp}-pid{pid}.log"

            file_handler = logging.FileHandler(log_path)
            file_handler.setFormatter(formatter)

            terminal_handler = logging.StreamHandler()
            terminal_handler.setFormatter(formatter)
            terminal_handler.setLevel(logging.ERROR)

            self._logger.addHandler(file_handler)
            self._logger.addHandler(terminal_handler)

    def start(self):
        try:
            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            try:
                self.server.bind((self.host, int(self.port)))
            except Exception as e:
                self._logger.info(
                    f"Error binding to port {self.port}: {e}, so binding to a random port"
                )
                self.server.bind((self.host, 0))
                self.port = self.server.getsockname()[1]

            self.server.listen()
            self.running = True

            self._logger.info(f"GDB Socket Server started on {self.host}:{self.port}")
            print(f"GDB Socket Server started on {self.host}:{self.port}")

            def set_pagination_off():
                gdb.execute("set pagination off ")

            gdb.post_event(set_pagination_off)
            self._logger.info("Set GDB pagination off")

            self.accept_thread = GdbThread(target=self._accept, daemon=True)
            self.accept_thread.start()

        except Exception as e:
            self._logger.error(f"Failed to start GDB Socket Server: {e}")

    def _accept(self):
        while self.running:
            try:
                client, address = self.server.accept()

                with self.clients_lock:
                    self.clients[address] = client

                self._logger.debug(
                    f"Accepted connection from {address}, total clients: {len(self.clients)}"
                )

                GdbThread(
                    target=self._process_requests,
                    args=(client, address),
                    daemon=True,
                ).start()

            except Exception as e:
                if self.running:
                    traceback.print_exc()
                    self._logger.error(f"Error accepting connection: {e}")

    def _process_requests_core(
        self, client: socket.socket, request: Request, status: PacketStatus
    ) -> None:
        try:
            async_exec = AsyncExec(request)

            # https://sourceware.org/gdb/current/onlinedocs/gdb.html/Threading-in-GDB.html
            # gdb.post_event is thread-safe, unlike gdb.execute.
            # gdb.post_event provides an ability to run any callable objects in the gdb main thread.
            gdb.post_event(async_exec)

            if status == PacketStatus.HAS_CALLBACK:
                self._logger.debug(
                    f"Posted event for callback {status} with {request}, waiting for completion"
                )
            else:
                self._logger.debug(
                    f"Posted event with {request}, waiting for completion"
                )

            # If the request does not has callback, we should send message to client immediately.
            # Because the client call must be not blocked.
            if status == PacketStatus.HAS_CALLBACK:
                socket_send(
                    client,
                    pickle.dumps(
                        (
                            Response(request.tag, request.tag),
                            PacketStatus.NO_CALLBACK,
                        )
                    ),
                    self._logger,
                )

            message = async_exec.get_result(timeout=DEFAULT_TIMEOUT)

            if isinstance(message, Exception):
                message = f"Error: {str(message)}"

            if status == PacketStatus.HAS_CALLBACK:
                self._logger.debug(f"Callback {status} with {request} completed")
            else:
                self._logger.debug(f"{request} completed")

            socket_send(
                client,
                pickle.dumps((Response(request.tag, message), status)),
                self._logger,
            )

        except Exception as e:
            traceback.print_exc()
            self._logger.error(f"Error running callback {status}: {e}")

    def _process_requests(self, client: socket.socket, address):
        while self.running:
            try:
                try:
                    data_bytes = socket_recv(client, self._logger)
                    payload: Tuple[Request, PacketStatus] = pickle.loads(data_bytes)
                except (TypeError, ValueError) as e:
                    # cloudpickle needs the same Python version to serialize/deserialize the object.
                    #
                    # Q: Why did we choose cloudpickle instead of pickle?
                    # A: Because standard pickle cannot deserialize objects
                    # that are not defined at the top level of a module.
                    #
                    # But Python standard pickle has backwards compatibility.
                    # So if standard pickle can deserialize an object,
                    # even if the object is not defined at the top level of a module,
                    # we can use standard pickle.
                    message = (
                        f"Error: {str(e)}\n"
                        f"maybe python version mismatch\n"
                        f"server python version: {sys.version}"
                    )
                    response = pickle.dumps(
                        (Response(0, message), PacketStatus.PYTHON_VERSION_MISMATCH)
                    )
                    socket_send(client, response, self._logger)
                    continue

                request, status = payload
                self._logger.info(f"Received request from {address}: {request}")
                assert isinstance(request, Request)

                GdbThread(
                    target=self._process_requests_core,
                    args=(client, request, status),
                    daemon=True,
                ).start()

            except ConnectionError:
                break
            except Exception as e:
                traceback.print_exc()
                self._logger.error(f"Error handling client {address}: {e}")

        try:
            client.close()
            self._logger.info(f"Closed connection from {address}")
            with self.clients_lock:
                if address in self.clients:
                    del self.clients[address]
                    self._logger.debug(
                        f"Removed client {address}, total clients: {len(self.clients)}"
                    )
        except Exception as e:
            self._logger.error(f"Error closing client socket {address}: {e}")

    def stop(self):
        with self.clients_lock:
            for address, client in self.clients.items():
                try:
                    client.close()
                    self._logger.info(f"Closed client connection from {address}")
                except Exception as e:
                    self._logger.error(f"Error closing client socket {address}: {e}")
            self.clients.clear()

        if self.running:
            try:
                self.server.close()
                self.running = False
            except Exception as e:
                self._logger.error(f"Error closing server socket: {e}")

        if self.accept_thread and self.accept_thread.is_alive():
            try:
                self.accept_thread.join(timeout=2.0)
                if self.accept_thread.is_alive():
                    self._logger.warning(
                        "Accept thread did not terminate within timeout"
                    )
            except Exception as e:
                self._logger.error(f"Error waiting for accept thread: {e}")

        self._logger.info("GDB Socket Server stopped")



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/tools/memory.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from typing import Optional

from mcp.server.fastmcp import Context

from ..utils import _exec_command


def register_memory_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def nxgdb_check_memory_leaks(
        ctx: Context, session_id: str, address: Optional[str] = None
    ) -> str:
        """Check for memory leaks in the target process.

        Analyzes allocated memory blocks to detect potential memory leaks.

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.
            address (Optional[str]): Memory address to analyze. When provided, dumps all
                valid nodes referenced by the memory block containing this address.

        Returns:
            str: Memory leak analysis results showing leaked blocks and their details.
        """
        command = "memleak"
        command += f" --address {address}" if address is not None else ""
        return await _exec_command(ctx, session_id, command)

    @gdb_mcp.tool()
    async def nxgdb_memalloc_inspect(
        ctx: Context,
        session_id: str,
        address: Optional[str] = None,
        log: Optional[str] = None,
        pid: Optional[int] = None,
        min_index: Optional[int] = None,
        max_index: Optional[int] = None,
        biggest: Optional[int] = None,
        top: Optional[int] = None,
        size: Optional[int] = None,
        orphan: Optional[bool] = None,
        no_pool: Optional[bool] = None,
        no_heap: Optional[bool] = None,
        no_group: Optional[bool] = None,
        no_backtrace: Optional[bool] = None,
        no_reverse: Optional[bool] = None,
        no_pid: Optional[int] = None,
        is_free: Optional[bool] = None,
        sort: Optional[str] = None,
    ) -> str:
        """Dump and analyze memory allocations in the target process.

        Executes the `memdump` command to provide detailed inspection of memory
        allocations and heap state. Supports extensive filtering and formatting
        options for customized memory analysis.

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.
            address (Optional[str]): Query which memory block contains this address.
                Example:
                `memdump --address 0x1234` to find the block containing address 0x1234.
            log (Optional[str]): Parse memory blocks from a log file generated by
                running memdump command on the device.
            pid (Optional[int]): Only dump memory allocated or freed by this specific
                task/process ID.
            min_index (Optional[int]): Minimum sequence number to include. Memory blocks
                outside the specified sequence range will be excluded.
            max_index (Optional[int]): Maximum sequence number to include. Memory blocks
                outside the specified sequence range will be excluded.
            biggest (Optional[int]): Sort by memory size (count * block size), equivalent
                to --sort size.
            top (Optional[int]): Display only the top N memory blocks.
            size (Optional[int]): Only dump memory blocks of this specific size.
            orphan (Optional[bool]): Only dump orphaned memory blocks (blocks without
                valid references).
            no_pool (Optional[bool]): Exclude pool memory, only show heap allocations.
            no_heap (Optional[bool]): Exclude heap memory, only show pool allocations.
            no_group (Optional[bool]): Disable memory block grouping. By default, blocks
                with same call stack, size, and PID are grouped together.
            no_backtrace (Optional[bool]): Skip printing memory block backtraces,
                significantly improving dump speed.
            no_reverse (Optional[bool]): Disable reverse sorting of results (default
                sorting is in reverse order).
            no_pid (Optional[int]): Exclude memory blocks with specific PID. Use -1
                to exclude mempool allocations.
            is_free (Optional[bool]): Dump free memory blocks instead of the default
                used memory blocks.
            sort (Optional[str]): Sort the dump results by specified criteria.
                Available options: 'size', 'nodesize', 'count', 'seq', 'address'.

        Returns:
            str: Detailed memory dump analysis showing allocation patterns, sizes,
                 backtraces, and other memory-related information.
        """
        command = "memdump"
        if address is not None:
            command += f" --address {address}"
        if log is not None:
            command += f" --log {log}"
        if pid is not None:
            command += f" --pid {pid}"
        if min_index is not None:
            command += f" --min {min_index}"
        if max_index is not None:
            command += f" --max {max_index}"
        if biggest is not None:
            command += " --biggest"
        if orphan is not None:
            command += " --orphan"
        if top is not None:
            command += f" --top {top}"
        if size is not None:
            command += f" --size {size}"
        if no_pool is not None:
            command += " --no-pool"
        if no_heap is not None:
            command += " --no-heap"
        if no_group is not None:
            command += " --no-group"
        if no_backtrace is not None:
            command += " --no-backtrace"
        if no_reverse is not None:
            command += " --no-reverse"
        if no_pid is not None:
            command += f" --no-pid {no_pid}"
        if is_free is not None:
            command += " --free"
        if sort is not None:
            command += f" --sort {sort}"
        return await _exec_command(ctx, session_id, command)

    @gdb_mcp.tool()
    async def nxgdb_find_memory_pattern(
        ctx: Context,
        session_id: str,
        pattern: str,
        memrange: Optional[str] = None,
        heap_only: Optional[bool] = None,
        globals_only: Optional[bool] = None,
    ) -> str:
        """Find memory addresses containing a specific pattern for memory corruption analysis.

        This tool searches for a specific pattern in memory to help analyze memory corruption
        issues. When memory has been corrupted/trampled with distinctive values, memfind can
        locate all occurrences of that pattern. Use with memdump -a <address> to investigate
        the code that might be causing the corruption.

        Common use case: When you detect memory corruption with a specific pattern/value,
        use memfind to locate all instances of that pattern in memory, then use memdump
        on those addresses to trace back to the source code that wrote those values.

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.
            pattern (str): Pattern to search for in memory. Can be a hex value (e.g., "0xdeadbeef"),
                string pattern, or byte sequence that you suspect is causing memory corruption.
            memrange (Optional[str]): Specific memory range to search within (e.g., "0x1000-0x2000").
                If not specified, searches the entire accessible memory space.
            heap_only (Optional[bool]): If True, restricts search to heap memory only,
                excluding stack and global variables.
            globals_only (Optional[bool]): If True, restricts search to global variables only,
                excluding heap and stack memory.

        Returns:
            str: List of memory addresses where the pattern was found, along with context
                 information that can be used with memdump for further investigation.
        """
        command = f"memfind {pattern}"

        if memrange is not None:
            command += f" --memrange {memrange}"
        if heap_only is not None:
            command += " --heap-only"
        if globals_only is not None:
            command += " --globals-only"

        return await _exec_command(ctx, session_id, command)

    @gdb_mcp.tool()
    async def nxgdb_fetch_memory_layout(
        ctx: Context,
        session_id: str,
        heap_only: Optional[bool] = None,
        globals_only: Optional[bool] = None,
    ) -> str:
        """Show RAM range of heap and sections for memory layout analysis.

        This tool displays comprehensive memory range information including all memory regions
        in the current system. It automatically parses ELF data sections and analyzes memory
        ranges within g_mmheap, providing a quick overview of memory regions for debugging.

        Key features:
        - Displays all system memory ranges
        - Automatically parses ELF data sections
        - Analyzes g_mmheap memory ranges
        - Provides quick memory region overview for debugging

        Use this tool to:
        - Get an overview of the memory layout
        - Understand heap and global variable locations
        - Prepare for targeted memory analysis with other tools

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.
            heap_only (Optional[bool]): If True, show only heap memory ranges,
                excluding other memory sections.
            globals_only (Optional[bool]): If True, show only global variable memory ranges,
                excluding heap and other sections.

        Returns:
            str: Detailed memory range information showing heap ranges, ELF data sections,
                 and other memory regions with their start/end addresses and properties.
        """
        command = "memrange"

        if heap_only is not None:
            command += " --heap-only"
        if globals_only is not None:
            command += " --globals-only"

        return await _exec_command(ctx, session_id, command)



############################################################################
# tools/pynuttx/nxgdb/tricore.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################
import argparse

import gdb

from . import autocompeletion, utils
from .backtrace import Backtrace

FCX_FREE_MASK = (0xFFFF << 0) | (0xF << 16)
PCXI_UL = 1 << 20
REG_LPC = 1
REG_UPC = 3

# Register offset in CSA

lower_offsets = [
    ("LPCXI", 0),
    ("LA11", 1),  # LPC
    ("A2", 2),
    ("A3", 3),
    ("D0", 4),
    ("D1", 5),
    ("D2", 6),
    ("D3", 7),
    ("A4", 8),
    ("A5", 9),
    ("A6", 10),
    ("A7", 11),
    ("D4", 12),
    ("D5", 13),
    ("D6", 14),
    ("D7", 15),
]
upper_offsets = [
    ("UPCXI", 0),
    ("PSW", 1),
    ("A10", 2),
    ("UA11", 3),  # UPC
    ("D8", 4),
    ("D9", 5),
    ("D10", 6),
    ("D11", 7),
    ("A12", 8),
    ("A13", 9),
    ("A14", 10),
    ("A15", 11),
    ("D12", 12),
    ("D13", 13),
    ("D14", 14),
    ("D15", 15),
]


def csa2addr(csa):
    return (csa & 0x000F0000) << 12 | (csa & 0x0000FFFF) << 6


class TricoreCSA:
    def __init__(self, addr, isupper=True):
        """
        Initialize TriCore CSA from memory address.
        Default is upper CSA, since tcb->xcp.regs points to upper CSA.

        :param addr: Memory address of the CSA
        :param upper: True for upper CSA, False for lower CSA

        """
        self.addr = addr
        self.isupper = isupper
        self.regs = {}
        for name, offset in upper_offsets if isupper else lower_offsets:
            try:
                self.regs[name] = utils.read_uint(addr + offset * 4)
            except Exception:
                print(
                    f"Error: failed to read register {name} at {hex(addr + offset * 4)}"
                )
                self.regs[name] = 0

    def dump_registers(self):
        """Dump all registers in the CSA."""
        line = ""
        for name, value in self.regs.items():
            line += f"{name}:0x{value:08X}  "
            if len(line) > 60:
                print(line)
                line = ""
        if line:
            print(line)

    def dump(self):
        """Dump CSA information."""
        print(self)
        self.dump_registers()

    @property
    def pc(self):
        """Get the PC value from the CSA."""
        return self.regs["UA11"] if self.isupper else self.regs["LA11"]

    @property
    def pcxi(self):
        """Get the PCXI value from the CSA."""
        return self.regs["UPCXI" if self.isupper else "LPCXI"]

    @property
    def next(self):
        """Get the next CSA from the current CSA."""
        pcxi = self.pcxi  # next CSA
        isupper = bool(pcxi & PCXI_UL)
        csa_addr = csa2addr(pcxi & FCX_FREE_MASK)
        return TricoreCSA(csa_addr, isupper) if csa_addr != 0 else None

    # iterator of all csa list from this one
    def __iter__(self):
        csa = self
        while csa:
            yield csa
            csa = csa.next

    # support using dict-like access
    def __getitem__(self, key):
        return self.regs[key]

    def __repr__(self):
        return f"{'Upper' if self.isupper else 'Lower'}CSA@{hex(self.addr)} PC=0x{self.pc:08X}"

    def __str__(self):
        return self.__repr__()


def csa_from_context(regs) -> TricoreCSA:
    """
    Get the TriCoreCSA from context regs.

    :param regs: The context registers address, usually from tcb->xcp.regs
    """
    # The CSA pointed by tcb->xcp.regs is always upper CSA
    # The lower CSA is stored above by one CSA size (0x40)
    # The lower CSA and upper CSA are linked together
    return TricoreCSA(int(regs) + 0x40, isupper=False)


def csa_from_tcb(tcb) -> TricoreCSA:
    """
    Get the TriCoreCSA from TCB.

    :param tcb: The TCB structure
    """
    if tcb is None:
        return None

    try:
        regs = int(tcb["xcp"]["regs"])
        # The CSA pointed by tcb->xcp.regs is always upper CSA
        return csa_from_context(regs)
    except gdb.error:
        return None


def csa_from_pcxi(pcxi) -> TricoreCSA:
    """
    Get the TriCoreCSA from PCXI value.

    :param pcxi: The PCXI value
    """
    isupper = bool(pcxi & PCXI_UL)
    csa_addr = csa2addr(pcxi & FCX_FREE_MASK)
    return TricoreCSA(csa_addr, isupper) if csa_addr != 0 else None


def dump_csa_chain(csa):
    """
    Dump the CSA chain starting from the given CSA.
    :param csa: The starting CSA
    """

    address = []
    for csa in list(csa):
        address.append(csa.pc)
        csa.dump()
    print(str(Backtrace(address)))


@autocompeletion.complete
class TricoreCSADump(gdb.Command):
    """Dump TriCore CSA list."""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-p",
            "--pid",
            type=int,
            default=None,
            help="PID of the task to dump CSA",
        )

        group.add_argument(
            "-a",
            "--addr",
            help="The context register address from TCB",
        )

        group.add_argument(
            "--pcxi",
            help="The PCXI register value to dump the CSA chain",
        )
        return parser

    def parse_argument(self, argv):
        try:
            return self.parser.parse_args(argv)
        except SystemExit:
            return None

    def __init__(self):
        arch = gdb.selected_inferior().architecture()
        if arch.name().startswith("TriCore"):
            super().__init__("tricore-dumpcsa", gdb.COMMAND_USER)
            self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_argument(gdb.string_to_argv(args))
        if args is None:
            return

        if args.pid is not None:
            csa = csa_from_tcb(utils.get_tcb(args.pid))
            dump_csa_chain(csa)
        elif args.pcxi is not None:
            pcxi = utils.parse_arg(args.pcxi)
            csa = csa_from_pcxi(pcxi)
            dump_csa_chain(csa)
        elif args.addr is not None:
            addr = utils.parse_arg(args.addr)
            csa = csa_from_context(addr)
            dump_csa_chain(csa)
        else:
            for tcb in utils.get_tcbs():
                print(
                    f"PID:{tcb['pid']}, state={tcb['task_state']}, regs={tcb['xcp']['regs']}"
                )
                csa = csa_from_tcb(tcb)
                dump_csa_chain(csa)



############################################################################
# tools/pynuttx/nxgdb/gdbrpc.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import logging
import os

import gdb
import gdbrpc
import psutil

from .server import Server

g_socket_server: Server = None


def start_gdb_socket_server(
    port: int = 20819, host: str = "localhost", log_level=logging.INFO
):
    """Start the GDB socket server"""
    global g_socket_server

    if g_socket_server and g_socket_server.running:
        print(
            f"GDB Socket Server already running on {g_socket_server.host}:{g_socket_server.port}"
        )
        return

    g_socket_server = Server(host, port, log_level)
    g_socket_server.start()


def stop_gdb_socket_server():
    """Stop the GDB socket server"""
    global g_socket_server

    if g_socket_server:
        g_socket_server.stop()
        g_socket_server = None


def get_gdb_socket_server_status():
    """Get server status"""
    if g_socket_server and g_socket_server.running:
        return {
            "running": True,
            "host": g_socket_server.host,
            "port": g_socket_server.port,
            "clients": len(g_socket_server.clients),
        }
    else:
        return {"running": False}


class StartSocketServer(gdb.Command):
    """Start GDB socket server command"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Start GDB socket server")
        parser.add_argument("--port", type=int, default=20819, help="Port to listen on")
        parser.add_argument("--host", default="localhost", help="Host to bind to")
        parser.add_argument(
            "--debug", action="store_true", default=False, help="Enable debug logging"
        )
        return parser

    def __init__(self):
        super().__init__("gdbrpc start", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def invoke(self, arg, from_tty):
        args = gdb.string_to_argv(arg)
        if not args:
            port = 20819
            host = "localhost"
            log_level = logging.INFO
        else:
            try:
                args = self.parser.parse_args(args)
                port = args.port
                host = args.host
                log_level = logging.DEBUG
            except SystemExit:
                return

        start_gdb_socket_server(port, host, log_level)


class StopSocketServer(gdb.Command):
    """Stop GDB socket server command"""

    def __init__(self):
        super().__init__("gdbrpc stop", gdb.COMMAND_USER)

    def invoke(self, arg, from_tty):
        stop_gdb_socket_server()


class SocketServerStatus(gdb.Command):
    """Get socket server status command"""

    def __init__(self):
        super().__init__("gdbrpc status", gdb.COMMAND_USER)

    @classmethod
    def get_memory_usage(cls):
        process = psutil.Process(os.getpid())
        memory_info = process.memory_info()
        return {
            "rss": memory_info.rss / 1024 / 1024,  # MB
            "vms": memory_info.vms / 1024 / 1024,  # MB
            "threads": len(process.threads()),
        }

    def invoke(self, arg, from_tty):
        status = get_gdb_socket_server_status()
        memory = self.get_memory_usage()

        if status["running"]:
            print("Socket Server Status:")
            print("  Running: Yes")
            print(f"  Host: {status['host']}")
            print(f"  Port: {status['port']}")
            print(f"  Connected Clients: {status['clients']}")
        else:
            print("Socket Server Status: Not running")
        print(f"  Memory Usage: {memory['rss']:.1f} MB RSS, {memory['vms']:.1f} MB VMS")
        print(f"  Active Threads: {memory['threads']}")


# !NOTE: This command is under development and may not work as expected
class StartSocketClient(gdb.Command):
    """Connect to GDB socket server"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Start GDB socket client")
        parser.add_argument(
            "--port", type=int, default=20819, help="Port to connect to"
        )
        parser.add_argument("--host", default="localhost", help="Host to connect to")
        return parser

    def __init__(self):
        super().__init__("gdbrpc connect", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def invoke(self, arg, from_tty):
        args = gdb.string_to_argv(arg)
        if not args:
            port = 20819
            host = "localhost"
        else:
            try:
                args = self.parser.parse_args(args)
                port = args.port
                host = args.host
            except SystemExit:
                return

        client = gdbrpc.ClientCLI(host, port)
        client.start()



############################################################################
# tools/pynuttx/nxgdb/thread.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
from enum import Enum
from typing import Union

import gdb
from nxelf.elf import LiefELF
from nxreg.register import Registers, get_arch_name

from . import autocompeletion, utils
from .stack import Stack

TCB_FLAG_TTYPE_SHIFT = 0
TCB_FLAG_TTYPE_MASK = 3 << TCB_FLAG_TTYPE_SHIFT
TCB_FLAG_TTYPE_TASK = 0 << TCB_FLAG_TTYPE_SHIFT
TCB_FLAG_TTYPE_PTHREAD = 1 << TCB_FLAG_TTYPE_SHIFT
TCB_FLAG_TTYPE_KERNEL = 2 << TCB_FLAG_TTYPE_SHIFT
TCB_FLAG_POLICY_SHIFT = 3
TCB_FLAG_POLICY_MASK = 3 << TCB_FLAG_POLICY_SHIFT
TCB_FLAG_SCHED_FIFO = 0 << TCB_FLAG_POLICY_SHIFT
TCB_FLAG_SCHED_RR = 1 << TCB_FLAG_POLICY_SHIFT
TCB_FLAG_SCHED_SPORADIC = 2 << TCB_FLAG_POLICY_SHIFT
TCB_FLAG_CPU_LOCKED = 1 << 5
TCB_FLAG_SIGNAL_ACTION = 1 << 6
TCB_FLAG_SYSCALL = 1 << 7
TCB_FLAG_EXIT_PROCESSING = 1 << 8
TCB_FLAG_FREE_STACK = 1 << 9
TCB_FLAG_HEAP_CHECK = 1 << 10
TCB_FLAG_HEAP_DUMP = 1 << 11
TCB_FLAG_DETACHED = 1 << 12
TCB_FLAG_FORCED_CANCEL = 1 << 13
TCB_FLAG_JOIN_COMPLETED = 1 << 14
TCB_FLAG_FREE_TCB = 1 << 15
TCB_FLAG_PREEMPT_SCHED = 1 << 16
TCB_FLAG_KILL_PROCESSING = 1 << 17

_SIGSET_NELEM = utils.get_field_nitems("struct sigset_t", "_elem")

CONFIG_SCHED_CPULOAD_NONE = utils.lookup_type("struct cpuload_s") is None


def is_thread_command_supported():
    # Check if the native thread command is available by compare the number of threads.
    # It should have at least CONFIG_SMP_NCPUS of idle threads.
    return len(gdb.selected_inferior().threads()) > utils.get_ncpus()


def get_task_state_desc(state):
    tstate = utils.enum("enum tstate_e")
    state = tstate(int(state))

    """
    Map task state enum to readable string, just like nxsched_get_stateinfo
    Avoid using nxsched_get_stateinfo in case it's not available.

    :param state: enum tstate_e
    :return: readable string
    """
    return {
        "TSTATE_TASK_INVALID": "Invalid",
        "TSTATE_TASK_PENDING": "Waiting,Unlock",
        "TSTATE_TASK_READYTORUN": "Ready",
        "TSTATE_TASK_ASSIGNED": "Assigned",
        "TSTATE_TASK_RUNNING": "Running",
        "TSTATE_TASK_INACTIVE": "Inactive",
        "TSTATE_WAIT_SEM": "Waiting,Semaphore",
        "TSTATE_WAIT_SIG": "Waiting,Signal",
        "TSTATE_WAIT_MQNOTEMPTY": "Waiting,MQ empty",
        "TSTATE_WAIT_MQNOTFULL": "Waiting,MQ full",
        "TSTATE_WAIT_PAGEFILL": "Waiting,Paging fill",
        "TSTATE_TASK_STOPPED": "Stopped",
    }.get(state.name, "Unknown")


class NxRegisters:
    saved_regs = None

    def __init__(self):
        self._registers = None

    @property
    def registers(self):
        if self._registers:
            return self._registers

        elf = gdb.objfiles()[0]
        elf = LiefELF(elf.filename)

        mapped_arch_name = get_arch_name()
        if not mapped_arch_name:
            raise ValueError("Architecture is not found in g_reg_table.\n")

        def read_memory(addr, size):
            return bytes(gdb.selected_inferior().read_memory(addr, size))

        self._registers = Registers(elf, arch=mapped_arch_name, readmem=read_memory)
        return self._registers

    def load(self, regs: Union[int, gdb.Value] = None):
        """Load registers from context register address"""
        self.registers.load(regs)
        for reg in self.registers:
            gdb.execute(f"set ${reg.name} = {reg.value}")

    def switch(self, pid):
        """Switch to the specified thread"""
        tcb = utils.get_tcb(pid)
        if not tcb:
            gdb.write(f"Thread {pid} not found\n")
            return

        if utils.task_is_running(tcb):
            # If the thread is running, then register is not in context but saved temporarily
            self.restore()
            return

        # Save current if this is the running thread, which is the case we never saved it before
        if not self.saved_regs:
            self.save()

        self.load(tcb["xcp"]["regs"])

    def save(self):
        """Save current registers"""
        if NxRegisters.saved_regs:
            # Already saved
            return

        registers = {}
        frame = gdb.newest_frame()
        for reg in self.registers:
            value = frame.read_register(reg.name)
            registers[reg.name] = value

        NxRegisters.saved_regs = registers

    def restore(self):
        if not NxRegisters.saved_regs:
            return

        for name, value in NxRegisters.saved_regs.items():
            gdb.execute(f"set ${name}={int(value)}")

        NxRegisters.saved_regs = None


g_registers = NxRegisters()


class RegInfoCommand(gdb.Command):
    """Display the register information"""

    def __init__(self):
        super().__init__("maintenance reginfo", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        registers = g_registers.registers
        header = ("Name", "Rmt Nr", "Offset", "Tcb Reg Off")
        print(
            "Name: the register name GDB uses.\n"
            "Rmt Nr: the register number in RSP packet, also the position in tcb.xcp.regs \n"
            "Tcb Reg Off: the byte offset in tcb.xcp.regs"
        )
        formatter = "{:<20} {:<10} {:<10} {:<10}"
        print(formatter.format(*header))
        for register in registers:
            print(
                formatter.format(
                    register.name,
                    register.regnum,
                    register.goffset,
                    register.toffset,
                )
            )


@autocompeletion.complete
class SetRegs(gdb.Command):
    """Load registers from TCB context memory address.
    Usage: setregs [regs]

    Etc: setregs
         setregs tcb->xcp.regs
         setregs g_pidhash[0]->xcp.regs

    Default to load from g_running_tasks if no args are provided.
    If the memory address is NULL, it will not set registers.
    """

    def get_argparser(self):
        parser = argparse.ArgumentParser(
            description="Set registers to the specified values"
        )

        parser.add_argument(
            "regs",
            nargs="?",
            default="",
            metavar="symbol",
            help="The memory address to load register values, use g_running_tasks.xcp.regs if not specified",
        )
        return parser

    def __init__(self):
        super().__init__("setregs", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        if args and args.regs:
            regs = utils.parse_and_eval(f"{args.regs}").cast(
                utils.lookup_type("char").pointer()
            )
        else:
            try:
                current_regs = utils.parse_and_eval("g_running_tasks[0].xcp.regs")
            except gdb.error as e:
                gdb.write(f"Failed to parse running tasks: {e}\n")
                return

            regs = current_regs.cast(utils.lookup_type("char").pointer())

        if regs == 0:
            gdb.write("regs is NULL\n")
            return

        g_registers.save()
        g_registers.load(regs)


class Nxinfothreads(gdb.Command):
    """Display information of all threads"""

    def __init__(self):
        super().__init__("info nxthreads", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        npidhash = utils.parse_and_eval("g_npidhash")
        pidhash = utils.parse_and_eval("g_pidhash")

        if utils.is_target_smp():
            gdb.write(
                "%-5s %-4s %-4s %-4s %-21s %-80s %-30s\n"
                % ("Index", "Tid", "Pid", "Cpu", "Thread", "Info", "Frame")
            )
        else:
            gdb.write(
                "%-5s %-4s %-4s %-21s %-80s %-30s\n"
                % ("Index", "Tid", "Pid", "Thread", "Info", "Frame")
            )

        for i, tcb in enumerate(utils.ArrayIterator(pidhash, npidhash)):
            if not tcb:
                continue

            pid = tcb["group"]["tg_pid"]
            tid = tcb["pid"]
            pc = utils.get_pc(tcb)
            thread = f"Thread {hex(tcb)}"
            index = f"*{i}" if utils.task_is_running(tcb) else f" {i}"

            statename = get_task_state_desc(tcb["task_state"])
            statename = f'\x1b{"[32;1m" if statename == "Running" else "[33;1m"}{statename}\x1b[m'

            if tcb["task_state"] == utils.parse_and_eval("TSTATE_WAIT_SEM"):
                mutex = tcb["waitobj"].cast(utils.lookup_type("sem_t").pointer())
                if utils.sem_is_mutex(mutex):
                    statename = f"Waiting,Mutex:{utils.mutex_get_holder(mutex)}"

            try:
                """Maybe tcb not have name member, or name is not utf-8"""
                info = (
                    "(Name: \x1b[31;1m%s\x1b[m, State: %s, Priority: %d, Stack: %d)"
                    % (
                        utils.get_task_name(tcb),
                        statename,
                        tcb["sched_priority"],
                        tcb["adj_stack_size"],
                    )
                )
            except gdb.error and UnicodeDecodeError:
                info = "(Name: Not utf-8, State: %s, Priority: %d, Stack: %d)" % (
                    statename,
                    tcb["sched_priority"],
                    tcb["adj_stack_size"],
                )

            line = gdb.find_pc_line(pc)
            if line.symtab:
                func = gdb.execute(f"info symbol {pc} ", to_string=True)
                frame = "\x1b[34;1m0x%x\x1b[\t\x1b[33;1m%s\x1b[m at %s:%d" % (
                    pc,
                    func.split()[0] + "()",
                    line.symtab,
                    line.line,
                )
            else:
                frame = "No symbol with pc"

            if utils.is_target_smp():
                cpu = f"{tcb['cpu']}"
                gdb.write(
                    "%-5s %-4s %-4s %-4s %-21s %-80s %-30s\n"
                    % (index, tid, pid, cpu, thread, info, frame)
                )
            else:
                gdb.write(
                    "%-5s %-4s %-4s %-21s %-80s %-30s\n"
                    % (index, tid, pid, thread, info, frame)
                )


class Nxthread(gdb.Command):
    """Switch to a specified thread"""

    def __init__(self):
        super().__init__("nxthread", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        npidhash = utils.parse_and_eval("g_npidhash")
        pidhash = utils.parse_and_eval("g_pidhash")
        arg = args.split(" ")
        arglen = len(arg)

        if arg[0] == "":
            pass
        elif arg[0] == "apply":
            if arglen <= 1:
                gdb.write("Please specify a thread ID list\n")
            elif arglen <= 2:
                gdb.write("Please specify a command following the thread ID list\n")

            elif arg[1] == "all":
                for i, tcb in enumerate(utils.ArrayIterator(pidhash, npidhash)):
                    if tcb == 0:
                        continue
                    try:
                        gdb.write(f"Thread {i} {tcb['name'].string()}\n")
                    except gdb.error and UnicodeDecodeError:
                        gdb.write(f"Thread {i}\n")

                    if not utils.task_is_running(tcb):
                        gdb.execute(f"setregs g_pidhash[{i}]->xcp.regs")

                    cmd_arg = ""
                    for cmd in arg[2:]:
                        cmd_arg += cmd + " "

                    gdb.write(gdb.execute(f"{cmd_arg}\n", to_string=True))
                    g_registers.restore()
            else:
                threadlist = []
                i = 0
                cmd = ""
                for i in range(1, arglen):
                    if arg[i].isnumeric():
                        threadlist.append(int(arg[i]))
                    else:
                        cmd += arg[i] + " "

                if len(threadlist) == 0 or cmd == "":
                    gdb.write("Please specify a thread ID list and command\n")
                else:
                    for i in threadlist:
                        if i >= npidhash:
                            break

                        if pidhash[i] == 0:
                            continue

                        try:
                            gdb.write(f"Thread {i} {pidhash[i]['name'].string()}\n")
                        except gdb.error and UnicodeDecodeError:
                            gdb.write(f"Thread {i}\n")

                        if not utils.task_is_running(utils.get_tcb(i)):
                            gdb.execute(f"setregs g_pidhash[{i}]->xcp.regs")

                        gdb.write(gdb.execute(f"{cmd}\n", to_string=True))
                        g_registers.restore()

        else:
            if (
                arg[0].isnumeric()
                and int(arg[0]) < npidhash
                and pidhash[int(arg[0])] != 0
            ):
                if utils.task_is_running(pidhash[int(arg[0])]):
                    g_registers.restore()
                else:
                    gdb.execute("setregs g_pidhash[%s]->xcp.regs" % arg[0])
            else:
                gdb.write(f"Invalid thread id {arg[0]}\n")


class Nxcontinue(gdb.Command):
    """Restore the registers and continue the execution"""

    def __init__(self):
        super().__init__("nxcontinue", gdb.COMMAND_USER)
        if not is_thread_command_supported():
            gdb.execute("define c\n nxcontinue \n end\n")
            gdb.write(
                "\n\x1b[31;1m if use thread command, please don't use 'continue', use 'c' instead !!!\x1b[m\n"
            )

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        g_registers.restore()
        gdb.execute("continue")


class Nxstep(gdb.Command):
    """Restore the registers and step the execution"""

    def __init__(self):
        super().__init__("nxstep", gdb.COMMAND_USER)
        if not is_thread_command_supported():
            gdb.execute("define s\n nxstep \n end\n")
            gdb.write(
                "\x1b[31;1m if use thread command, please don't use 'step', use 's' instead !!!\x1b[m\n"
            )

    def invoke(self, args, from_tty):
        g_registers.restore()
        gdb.execute("step")


class TaskType(Enum):
    TASK = 0
    PTHREAD = 1
    KTHREAD = 2


class TaskSchedPolicy(Enum):
    FIFO = 0
    RR = 1
    SPORADIC = 2


class Ps(gdb.Command):
    def __init__(self):
        super().__init__("ps", gdb.COMMAND_USER)
        self._fmt_wxl = "{0: <{width}}"
        # By default we align to the right, whcih respects the nuttx foramt
        self._fmt_wx = "{0: >{width}}"
        self._char_ptr_ptr_type = None
        self._mutex_t_ptr_type = None
        self._sem_t_ptr_type = None
        self._pthread_tcb_s_ptr_type = None

    def get_cached_type(self, type_name):
        if type_name == "char_ptr_ptr" and not self._char_ptr_ptr_type:
            self._char_ptr_ptr_type = gdb.lookup_type("char").pointer().pointer()
        elif type_name == "mutex_t_ptr" and not self._mutex_t_ptr_type:
            self._mutex_t_ptr_type = utils.lookup_type("mutex_t").pointer()
        elif type_name == "sem_t_ptr" and not self._sem_t_ptr_type:
            self._sem_t_ptr_type = utils.lookup_type("sem_t").pointer()
        elif type_name == "pthread_tcb_s_ptr" and not self._pthread_tcb_s_ptr_type:
            self._pthread_tcb_s_ptr_type = utils.lookup_type(
                "struct pthread_tcb_s"
            ).pointer()

        return getattr(self, f"_{type_name}_type")

    def parse_and_show_info(self, tcb):
        def eval2str(cls, x):
            return cls(int(x)).name

        pid = int(tcb["pid"])
        group = int(tcb["group"]["tg_pid"])
        priority = int(tcb["sched_priority"])
        flags = int(tcb["flags"])

        policy = eval2str(
            TaskSchedPolicy,
            (flags & TCB_FLAG_POLICY_MASK) >> TCB_FLAG_POLICY_SHIFT,
        )

        task_type = eval2str(
            TaskType,
            (flags & TCB_FLAG_TTYPE_MASK) >> TCB_FLAG_TTYPE_SHIFT,
        )

        npx = "P" if (flags & TCB_FLAG_EXIT_PROCESSING) else "-"

        waiter = ""
        if tcb["waitobj"]:
            waitobj = tcb["waitobj"].cast(self.get_cached_type("sem_t_ptr"))
            if utils.sem_is_mutex(waitobj):
                mutex = tcb["waitobj"].cast(utils.lookup_type("sem_t").pointer())
                waiter = str(utils.mutex_get_holder(mutex))

        state_and_event = get_task_state_desc(int(tcb["task_state"]))
        if waiter:
            state_and_event += "@MutexHolder: " + waiter
        state_and_event = state_and_event.split(",")

        # Append a null str here so we don't need to worry
        # about the number of elements as we only want the first two
        state, event = (
            state_and_event if len(state_and_event) > 1 else state_and_event + [""]
        )

        sigmask = "{0:#0{1}x}".format(
            sum(int(tcb["sigprocmask"]["_elem"][i] << i) for i in range(_SIGSET_NELEM)),
            _SIGSET_NELEM * 8 + 2,
        )[
            2:
        ]  # exclude "0x"

        st = Stack(
            int(tcb["stack_base_ptr"]),
            int(tcb["adj_stack_size"]),
            utils.get_sp(tcb),
        )

        stacksz = st._stack_size
        used = st.max_usage()
        filled = "{0:.2%}".format(used / stacksz)

        cpu = int(tcb["cpu"]) if utils.is_target_smp() else 0

        # For a task we need to display its cmdline arguments, while for a thread we display
        # pointers to its entry and argument
        cmd = ""
        name = utils.get_task_name(tcb)

        if (flags & TCB_FLAG_TTYPE_MASK) == TCB_FLAG_TTYPE_PTHREAD:
            entry = tcb["entry"]["main"]
            ptcb = tcb.cast(utils.lookup_type("struct pthread_entry_s").pointer())
            arg = ptcb["arg"]
            cmd = f"{name} {hex(entry)} {hex(arg)}"
        elif pid < utils.get_ncpus():
            cmd = name
        else:
            # For tasks other than pthreads, hence need to get its command line
            # arguments from
            tls_info_type = utils.lookup_type("struct tls_info_s").pointer()
            stack_ptr = tcb["stack_alloc_ptr"].cast(tls_info_type)
            argv = tcb["stack_alloc_ptr"] + stack_ptr["tl_size"]
            args = []
            parg = argv.cast(self.get_cached_type("char_ptr_ptr")) + 1
            while parg.dereference():
                args.append(parg.dereference().string())
                parg += 1
            cmd = " ".join([name] + args)

        if not CONFIG_SCHED_CPULOAD_NONE:
            g_cpuload_total = int(utils.parse_and_eval("g_cpuload_total"))
            load = "{0:.1%}".format(
                int(tcb["ticks"]) / g_cpuload_total if g_cpuload_total else 0
            )
        else:
            load = "Dis."

        if hasattr(self, "table"):
            self.table.add_row(
                [
                    pid,
                    group,
                    cpu,
                    priority,
                    policy,
                    task_type,
                    npx,
                    state,
                    event,
                    sigmask,
                    stacksz,
                    used,
                    filled,
                    load,
                    cmd,
                ]
            )
        else:
            gdb.write(
                " ".join(
                    (
                        self._fmt_wx.format(pid, width=5),
                        self._fmt_wx.format(group, width=5),
                        self._fmt_wx.format(cpu, width=3),
                        self._fmt_wx.format(priority, width=3),
                        self._fmt_wxl.format(policy, width=8),
                        self._fmt_wxl.format(task_type, width=7),
                        self._fmt_wx.format(npx, width=3),
                        self._fmt_wxl.format(state, width=8),
                        self._fmt_wxl.format(event, width=9),
                        self._fmt_wxl.format(sigmask, width=8),
                        self._fmt_wx.format(stacksz, width=7),
                        self._fmt_wx.format(used, width=7),
                        self._fmt_wx.format(filled, width=6),
                        self._fmt_wx.format(load, width=6),
                        cmd,
                    )
                )
            )
            gdb.write("\n")

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        prettytable = utils.import_check(
            "prettytable",
            errmsg="Execute `pip install prettytable` for better printing result.\n",
        )
        if prettytable:
            self.table = prettytable.PrettyTable()
            self.table.align = "l"
            self.table.field_names = [
                "PID",
                "GROUP",
                "CPU",
                "PRI",
                "POLICY",
                "TYPE",
                "NPX",
                "STATE",
                "EVENT",
                "SIGMASK",
                "STACK",
                "USED",
                "FILLED",
                "LOAD",
                "COMMAND",
            ]
        else:
            gdb.write(
                " ".join(
                    (
                        self._fmt_wx.format("PID", width=5),
                        self._fmt_wx.format("GROUP", width=5),
                        self._fmt_wx.format("CPU", width=3),
                        self._fmt_wx.format("PRI", width=3),
                        self._fmt_wxl.format("POLICY", width=8),
                        self._fmt_wxl.format("TYPE", width=7),
                        self._fmt_wx.format("NPX", width=3),
                        self._fmt_wxl.format("STATE", width=8),
                        self._fmt_wxl.format("EVENT", width=9),
                        self._fmt_wxl.format("SIGMASK", width=_SIGSET_NELEM * 8),
                        self._fmt_wx.format("STACK", width=7),
                        self._fmt_wx.format("USED", width=7),
                        self._fmt_wx.format("FILLED", width=3),
                        self._fmt_wx.format("LOAD", width=6),
                        "COMMAND",
                    )
                )
            )
            gdb.write("\n")

        for tcb in utils.get_tcbs():
            try:
                self.parse_and_show_info(tcb)
            except gdb.error as e:
                gdb.write(f"[Error] GDB error while processing TCB: {e}\n")
            except Exception as e:
                gdb.write(f"[Error] Unexpected error: {e}\n")
        if hasattr(self, "table"):
            gdb.write(f"{self.table.get_string()}\n")

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Thread Information",
            "summary": "Thread information",
            "command": "ps",
            "result": "info",
            "category": utils.DiagnoseCategory.sched,
            "message": gdb.execute("ps", to_string=True),
        }



############################################################################
# tools/pynuttx/tests/test_mock_utils.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import unittest
from unittest.mock import MagicMock, patch

import gdb
from nxgdb import utils

# Loading gdb/__init__.py will append the parent directory to sys.path
# Hence we don't need to do importing manually here


class TestGDBEvalOrNone(unittest.TestCase):
    """
    TODO
    """

    pass


class TestGetSymbolValue(unittest.TestCase):
    """
    TODO
    """

    pass


class TestIsDecimal(unittest.TestCase):
    """
    TODO
    """

    pass


class TestIsHexDecimal(unittest.TestCase):
    """
    TODO
    """

    pass


class TestHexdump(unittest.TestCase):
    """
    TODO
    """

    pass


class TestGetTargetEndianness(unittest.TestCase):
    """
    TODO
    """

    pass


class TestReadBinaryData(unittest.TestCase):
    """
    TODO
    """

    pass


class TestIsTargetArch(unittest.TestCase):
    def setUp(self):
        # Mock the gdb.newest_frame().architecture() method
        gdb.newest_frame = MagicMock()
        self._mock_frame = gdb.Frame = MagicMock()
        self._mock_gdb_exec = gdb.execute = MagicMock()
        self._mock_arch_name = gdb.newest_frame().architecture().name

    def tearDown(self):
        self._mock_frame.reset_mock()
        self._mock_gdb_exec.reset_mock()
        self._mock_arch_name.reset_mock()
        gdb.newest_frame.reset_mock()

    def test_is_target_arch_exact_match(self):
        self._mock_arch_name.return_value = "i386:x86_64"
        # Test if the test() function returns True
        self.assertTrue(utils.is_target_arch("i386:x86_64", exact=True))
        self.assertFalse(utils.is_target_arch("x86", exact=True))

    def test_is_target_arch_non_exact_match(self):
        self._mock_arch_name.return_value = "i386:x86_64"
        # Test if the test() function returns True
        self.assertTrue(utils.is_target_arch("x86"))
        self.assertFalse(utils.is_target_arch("arm"))

    @patch("nxgdb.utils.target_arch", None)
    def test_is_target_arch_no_attr_architecture_auto(self):
        del self._mock_frame.architecture

        self._mock_gdb_exec.return_value = (
            'The target architecture is set to "auto" (currently "armv7e-m").'
        )
        self.assertTrue(utils.is_target_arch("arm"))

    @patch("nxgdb.utils.target_arch", None)
    def test_is_target_arch_no_attr_architecture_specific(self):
        del self._mock_frame.architecture

        self._mock_gdb_exec.return_value = 'The target architecture is set to "riscv"'
        self.assertTrue(utils.is_target_arch("riscv", exact=True))


class TestIsTargetSMP(unittest.TestCase):
    """
    TODO
    """

    pass


class TestInInterruptContext(unittest.TestCase):
    """
    TODO
    """

    pass


class TestGetArchRegister(unittest.TestCase):
    """
    TODO
    """

    pass


class TestGetTcbs(unittest.TestCase):
    """
    TODO
    """

    pass



############################################################################
# tools/pynuttx/nxgdb/dmesg.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb

from . import utils

CONFIG_RAMLOG_SYSLOG = utils.get_static_symbol("g_ramlog_channel_ops")
CONFIG_SYSLOG_RPMSG = utils.get_static_symbol("g_rpmsg_channel_ops")


class Dmesg(gdb.Command):
    def __init__(self):
        if CONFIG_RAMLOG_SYSLOG or CONFIG_SYSLOG_RPMSG:
            super().__init__("dmesg", gdb.COMMAND_USER)

    def _get_ramlog(self):
        if not (sysdev := utils.gdb_eval_or_none("g_sysdev")):
            return "RAM log not available"

        rl_header = sysdev["rl_header"]
        rl_bufsize = sysdev["rl_bufsize"]

        offset = rl_header["rl_head"] % rl_bufsize  # Currently writing to this offset
        tail = rl_bufsize - offset  # Total size till buffer end.
        rl_buffer = int(rl_header["rl_buffer"].address)  # rl_buffer is a char array

        inf = gdb.selected_inferior()
        buf = bytes(inf.read_memory(offset + rl_buffer, tail) if tail else b"")
        buf = buf.lstrip(b"\0")  # Remove leading NULLs
        buf += bytes(inf.read_memory(rl_buffer, offset) if offset else b"")
        buf = buf.replace(
            b"\0", "␀".encode("utf-8")
        )  # NULL is valid utf-8 but not printable

        return buf.decode("utf-8", errors="replace")

    def _get_rpmsg_syslog(self):
        if not (priv := utils.gdb_eval_or_none("g_syslog_rpmsg")):
            return "RPMsg syslog not avaliable"

        inf = gdb.selected_inferior()
        buffer = bytes(inf.read_memory(priv.buffer, priv.size) if priv.size else b"")
        buf = buffer.replace(b"\0", "␀".encode("utf-8"))
        return buf.decode("utf-8", errors="replace")

    def diagnose(self, *args, **kwargs):
        ramlog = self._get_ramlog()
        rpmsg_syslog = self._get_rpmsg_syslog()

        return {
            "title": "RAM log and RPMsg Syslog",
            "summary": (
                f"RAM log length: {len(ramlog)} bytes. RPMSG log length:{len(rpmsg_syslog)} bytes."
            ),
            "result": "info",
            "category": utils.DiagnoseCategory.system,
            "command": "dmesg",
            "message": f"RAM log:\n{ramlog}\n RPMSG syslog:{rpmsg_syslog}",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        ramlog = self._get_ramlog()
        rpmsg_syslog = self._get_rpmsg_syslog()

        if CONFIG_RAMLOG_SYSLOG:
            print(f"RAM log:{ramlog}\n---END of RAMLOG")

        if CONFIG_SYSLOG_RPMSG:
            print(f"RPMSG syslog:{rpmsg_syslog}\n---END of RPMSG SYSLOG---")



############################################################################
# tools/pynuttx/nxgdb/fs/command.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb
import nxgdb.autocompeletion as autocompeletion
import nxgdb.fs.fatfs as fatfs
import nxgdb.fs.littlefs as littlefs
import nxgdb.fs.lrofs as lrofs
import nxgdb.fs.romfs as romfs
import nxgdb.fs.yaffs as yaffs
import nxgdb.utils as utils
from nxgdb.fs.fs import print_fdinfo_by_tcb
from nxgdb.fs.inode import (
    Inode,
    InodeType,
    foreach_inode,
    fstype_filter,
    get_fstype,
    get_root_inode,
    inode_gettype,
    print_inode_info,
)

from .utils import (
    CONFIG_DISABLE_MOUNTPOINT,
    CONFIG_FS_FATFS,
    CONFIG_FS_LITTLEFS,
    CONFIG_FS_LROFS,
    CONFIG_FS_ROMFS_CACHE_NODE,
    CONFIG_FS_SHMFS,
    CONFIG_FS_YAFFS,
)


@autocompeletion.complete
class ForeachInode(gdb.Command):
    """Dump each inode info"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="foreach inode command")
        parser.add_argument(
            "-L",
            "--level",
            type=int,
            default=4096,
            help="Only render the tree to a specific depth",
        )
        parser.add_argument(
            "--nodetype",
            type=str,
            choices=[e.name.lower() for e in InodeType],
            default=None,
            help="Only show the specific type of inode",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="Show more information",
        )
        parser.add_argument(
            "addr_or_expr",
            type=str,
            metavar="symbol",
            nargs="?",
            default=None,
            help="set the start inode to be tranversed",
        )
        return parser

    def __init__(self):
        super().__init__("foreach inode", gdb.COMMAND_USER)
        utils.alias("inode-foreach", "foreach inode")
        self.parser = self.get_argparser()

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("foreach inode", to_string=True)

        return {
            "title": "File Node Information",
            "summary": "inode formation dump",
            "command": "foreach inode",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "message": output,
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(args.split(" "))
        if not args:
            return

        root = (
            get_root_inode(args.addr_or_expr)
            if args.addr_or_expr
            else utils.gdb_eval_or_none("g_root_inode")
        )

        nodetype = InodeType[args.nodetype.upper()] if args.nodetype else None
        if nodetype:
            self.nodes = list(
                filter(lambda x: inode_gettype(x[0]) == nodetype, foreach_inode(root))
            )

        print_inode_info(
            self.nodes if hasattr(self, "nodes") else None,
            Inode(root),
            maxlevel=args.level,
            type=nodetype,
            verbose=args.verbose,
        )


@autocompeletion.complete
class Fdinfo(gdb.Command):
    """Dump fd info information of process"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(
            description="Get fdinfo for a process or all processes."
        )
        parser.add_argument("-p", "--pid", type=int, help="Optional process ID")
        return parser

    def __init__(self):
        super().__init__("fdinfo", gdb.COMMAND_DATA)
        self.total_fd_count = 0
        self.parser = self.get_argparser()

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("fdinfo", to_string=True)

        return {
            "title": "File Descriptors Information",
            "summary": f"Total files opened:{self.total_fd_count}",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "command": "fdinfo",
            "message": output,
        }

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            gdb.write("Invalid arguments.\n")
            return

        self.processed_groups = set()
        self.total_fd_count = 0
        tcbs = [utils.get_tcb(args.pid)] if args.pid else utils.get_tcbs()
        for tcb in tcbs:
            print_fdinfo_by_tcb(tcb, self.processed_groups, self.total_fd_count)


class Mount(gdb.Command):
    def __init__(self):
        if not CONFIG_DISABLE_MOUNTPOINT:
            super().__init__("mount", gdb.COMMAND_USER)
            self.mount_count = 0

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("mount", to_string=True)

        return {
            "title": "File System Mount Information",
            "summary": f"Total {self.mount_count} mount points",
            "command": "mount",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "message": output or "No mount",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        self.mount_count = 0
        nodes = filter(
            lambda x: inode_gettype(x[0]) == InodeType.MOUNTPT, foreach_inode()
        )
        for node, path in nodes:
            fstype = get_fstype(node)
            gdb.write("  %s type %s\n" % (path, fstype))
            self.mount_count += 1


class InfoShmfs(gdb.Command):
    """Show share memory usage"""

    def __init__(self):
        if CONFIG_FS_SHMFS:
            super().__init__("info shm", gdb.COMMAND_USER)
            self.total_size = 0
            self.block_count = 0

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("info shm", to_string=True)

        return {
            "title": "Share Memory Usage",
            "summary": f"Total used:{self.total_size}kB, {self.block_count}blocks",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "command": "info shm",
            "message": output or "No InfoShmfs",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        self.total_size = 0
        self.block_count = 0
        nodes = filter(lambda x: inode_gettype(x[0]) == InodeType.SHM, foreach_inode())
        for node, path in nodes:
            obj = node.i_private.cast(
                gdb.lookup_type("struct shmfs_object_s").pointer()
            )
            length = obj.length
            paddr = obj.paddr
            print(f"  {path} memsize: {length}, paddr: {paddr}")

            self.total_size += length / 1024
            self.block_count += 1


@autocompeletion.complete
class InfoRomfs(gdb.Command):
    """Show romfs cache information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-P",
            "--path",
            type=str,
            default=None,
            help="set the romfs path to be dumped",
        )
        return parser

    def __init__(self):
        if CONFIG_FS_ROMFS_CACHE_NODE:
            super().__init__("info romfs", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("info romfs", to_string=True)

        return {
            "title": "Romfs Cache Information",
            "summary": "Romfs nodeinfo dump",
            "command": "info romfs",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "message": output or "No romfs information",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(gdb.string_to_argv(args))
        nodes = filter(fstype_filter("romfs"), foreach_inode())
        for node, path in nodes:
            if args and args.path and path != args.path:
                continue
            romfs.dump_romfs_cache(node, path)


@autocompeletion.complete
class InfoLittlefs(gdb.Command):
    """Show littlefs cache information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-P",
            "--path",
            type=str,
            default=None,
            help="set the littlefs path to be dumped",
        )
        return parser

    def __init__(self):
        if CONFIG_FS_LITTLEFS:
            super().__init__("info littlefs", gdb.COMMAND_USER)
            self.parser = self.get_argparser()

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("info littlefs", to_string=True)

        return {
            "title": "Littlefs cache information",
            "summary": "Littlefs nodeinfo dump",
            "command": "info littlefs",
            "result": "info",
            "message": output or "No littlefs information",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(gdb.string_to_argv(args))
        nodes = filter(fstype_filter("littlefs"), foreach_inode())
        for node, path in nodes:
            if args and args.path and path != args.path:
                continue
            littlefs.dump_little_cache(node, path)


@autocompeletion.complete
class InfoYaffs(gdb.Command):
    """Show yaffs cache information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=gdb.__doc__)
        parser.add_argument(
            "-P",
            "--path",
            type=str,
            default=None,
            help="set the yaffs path to be dumped",
        )
        return parser

    def __init__(self):
        if CONFIG_FS_YAFFS:
            super().__init__("info yaffs", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("info yaffs", to_string=True)
        return {
            "title": "Yaffs Cache Information",
            "summary": "Yaffs information dump",
            "command": "info Yaffs",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "message": output or "No yaffs information",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(gdb.string_to_argv(args))
        nodes = filter(fstype_filter("yaffs"), foreach_inode())
        for node, path in nodes:
            if args and args.path and path != args.path:
                continue
            yaffs.dump_yaffs_cache(node, path)


class InfoFatfs(gdb.Command):
    """Show fatfs information"""

    def __init__(self):
        if CONFIG_FS_FATFS:
            super().__init__("info fatfs", gdb.COMMAND_USER)

    def parse_arguments(self, argv):
        parser = argparse.ArgumentParser(description=gdb.__doc__)
        parser.add_argument(
            "-P",
            "--path",
            type=str,
            default=None,
            help="set the fatfs path to be dumped",
        )
        parser.add_argument(
            "-F",
            "--filep",
            type=str,
            default=None,
            help="Show file information",
        )
        try:
            args = parser.parse_args(argv)
        except SystemExit:
            return None
        return args

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("info fatfs", to_string=True)
        return {
            "title": "Fatfs information",
            "summary": "Fatfs information dump",
            "command": "info fatfs",
            "result": "info",
            "message": output or "No fatfs information",
        }

    def invoke(self, args, from_tty):
        args = self.parse_arguments(gdb.string_to_argv(args))
        nodes = filter(fstype_filter("fatfs"), foreach_inode())
        if args and args.filep:
            fatfs.dump_fatfs_file(utils.Value(utils.parse_arg(args.filep)))
            return
        for node, path in nodes:
            if args and args.path and path != args.path:
                continue
            fatfs.dump_fatfs_cache(node, path)


class InfoLrofs(gdb.Command):
    """Show lrofs cache information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-P",
            "--path",
            type=str,
            default=None,
            help="set the lrofs path to be dumped",
        )
        return parser

    def __init__(self):
        if CONFIG_FS_LROFS:
            super().__init__("info lrofs", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    def diagnose(self, *args, **kwargs):
        output = gdb.execute("info lrofs", to_string=True)

        return {
            "title": "Lrofs Cache Information",
            "summary": "Lrofs nodeinfo dump",
            "command": "info lrofs",
            "result": "info",
            "category": utils.DiagnoseCategory.fs,
            "message": output or "No lrofs information",
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(gdb.string_to_argv(args))
        nodes = filter(fstype_filter("lrofs"), foreach_inode())
        for node, path in nodes:
            if args and args.path and path != args.path:
                continue
            lrofs.dump_lrofs_cache(node, path)



###########################################################################
# tools/pynuttx/nxgdb/target.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import multiprocessing
import socket

import gdb
import nxstub
from nxreg.register import get_arch_name

from . import autocompeletion, utils


def get_unused_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("127.0.0.1", port))
        except OSError:
            return True
        return False


@autocompeletion.complete
class Target(gdb.Command):
    """Use nxstub to parse crash log dump, core dump or memory dump, as target."""

    parser = nxstub.get_argparser()

    def __init__(self):
        super().__init__("target stub", gdb.COMMAND_USER)
        utils.alias("target nxstub", "target stub")
        self.process = None

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        self.dont_repeat()

        if "-e" not in args and "--elf" not in args:
            args += f" -e {gdb.objfiles()[0].filename}"

        arch = get_arch_name()  # Convert to nxstub arch name

        if "-a" not in args and "--arch" not in args:
            args += f" -a {arch}"
        else:
            print(f"Hint: no need to specify architecture, current arch: {arch}")

        args = gdb.string_to_argv(args)
        try:
            parsed = nxstub.parse_args(args)
        except SystemExit:
            return

        if parsed.arch != arch:
            print(f"Warning: current arch {arch} does not match nxstub {parsed.arch}")

        if is_port_in_use(parsed.port):
            print(f"Port {parsed.port} is already in use, try to use another port.")
            parsed.port = get_unused_port()
            print(f"Use port {parsed.port} instead.")

        if parsed.timeout:
            gdb.execute(f"set tcp connect-timeout {parsed.timeout}", from_tty=True)
            print(f"Set GDB timeout: {parsed.timeout}")

        # If currently has connection to target, disconnect it
        if utils.check_inferior_valid():
            gdb.execute("detach", from_tty=True)

        def kill(event=None):
            if self.process:
                self.process.kill()
                self.process.join()
                self.process = None
                print("nxstub process killed")

        if self.process:
            y = input("nxstub process already running, kill it? [y/n] ")
            if y.lower() != "y":
                return

            kill()

        def stub_main(args):
            import signal

            # Ignore the Ctrl+C signal
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            nxstub.main(args)

        process = multiprocessing.Process(target=stub_main, args=(parsed,))
        process.start()
        self.process = process

        # Check if the process exits immediately
        process.join(timeout=1)
        if not process.is_alive():
            self.process = None
            return

        gdb.events.gdb_exiting.connect(kill)
        print("")

        # Wait server to start
        gdb.execute(f"target remote :{parsed.port}", from_tty=True)



############################################################################
# tools/pynuttx/gdbinit.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import sys
from os import path
from pathlib import Path

here = path.dirname(path.abspath(__file__))

nuttx_path = Path(__file__).resolve().parents[2]
libcxx_path = path.join(
    nuttx_path, "libs", "libxx", "libcxx", "libcxx", "utils", "gdb", "libcxx"
)

if __name__ == "__main__":
    if here not in sys.path:
        sys.path.insert(0, here)

    if path.exists(libcxx_path):
        sys.path.insert(0, libcxx_path)
        import gdb
        import printers

        printer = printers.LibcxxPrettyPrinter("libcxx")
        gdb.pretty_printers.append(printer)
        printers.register_libcxx_printer_loader()

    modules = ("gdbrpc", "nxelf", "nxgdb", "nxreg", "nxstub", "nxtrace")
    for key in list(sys.modules):
        if key.startswith(modules):
            del sys.modules[key]

    import nxgdb  # noqa: F401

    print("GDB Plugin Loaded Successfully")



############################################################################
# tools/pynuttx/nxgdb/venv.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import subprocess
import sys
import venv
from os import path
from typing import Optional

import gdb

from . import autocompeletion

venv_dir = ".gdbenv"
venv_install_flag = "not initialized"


def install_dependency(
    location: str = venv_dir,
    requirements: Optional[str] = None,
    package: Optional[str] = None,
) -> None:
    if requirements:
        if not path.exists(requirements):
            raise RuntimeError(f"requirements.txt not found at {requirements}")
    elif package:
        if not package.strip():
            raise RuntimeError("Package name is empty")
    else:
        raise RuntimeError("Either requirements or package must be specified")

    pip_path = path.join(location, "bin", "pip")
    if not path.exists(pip_path):
        raise RuntimeError(f"pip not found at {pip_path}, please create venv first")

    if requirements:
        pip_cmd = [pip_path, "install", "-r", requirements]
    else:
        pip_cmd = [pip_path, "install", package]

    try:
        subprocess.check_call(pip_cmd)
    except KeyboardInterrupt:
        print("pip install interrupted by user. You can re-run to resume.")
    except subprocess.CalledProcessError as e:
        print(f"pip install failed (exit {e.returncode}). re-run to retry.")
    else:
        with open(venv_install_flag, "w") as f:
            f.write("s")


def init_venv_and_install_dependency(location: str = venv_dir) -> None:
    requirements = path.join(
        path.dirname(path.abspath(__file__)), "../requirements.txt"
    )

    if path.exists(location):
        print(f"Virtual environment already exists at {location}")
    else:
        print(f"Creating virtual environment at {location}")
        venv.EnvBuilder(with_pip=True, symlinks=True).create(location)

    if not path.exists(venv_install_flag):
        install_dependency(location=location, requirements=requirements)


@autocompeletion.complete
class GDBVenv(gdb.Command):
    """Create NuttX GDB Plugin virtual environment"""

    def __init__(self):
        super().__init__("gdbvenv", gdb.COMMAND_USER)
        self.parser = self.get_argparser()
        self.use_venv()  # If venv already exists, use it immediately

    def get_argparser(self):
        parser = argparse.ArgumentParser(
            description="GDB Plugin Virtual Environment Manager",
        )
        parser.add_argument(
            "venvdir",
            type=str,
            default=venv_dir,
            nargs="?",
            help="Path to the virtual environment directory",
        )
        parser.add_argument(
            "--install",
            type=str,
            help="Path to requirements.txt or package, or package name to install into the virtual environment",
        )
        return parser

    def use_venv(self, location: str = venv_dir) -> None:
        py_version = f"python{sys.version_info.major}.{sys.version_info.minor}"
        site_packages = path.join(location, "lib", py_version, "site-packages")
        if site_packages in sys.path:
            return

        if path.exists(site_packages):
            print(f"Adding site-packages: {site_packages}")
            sys.path.insert(0, site_packages)
        else:
            print(f"site-packages not found at {site_packages}")

    def invoke(self, arg, from_tty):  # type: ignore
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        global venv_install_flag
        venv_install_flag = path.join(args.venvdir, ".gdbenv_installed")

        if args.install:
            if args.install.endswith("requirements.txt"):
                install_dependency(location=args.venvdir, requirements=args.install)
            else:
                install_dependency(location=args.venvdir, package=args.install)
        else:
            init_venv_and_install_dependency(location=args.venvdir)

        self.use_venv(location=args.venvdir)
